# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
import random
import shutil
import xml.etree.ElementTree as ET

from .....utils.deps import function_requires_deps, is_dep_available
from .....utils.errors import ConvertFailedError
from .....utils.file_interface import custom_open, write_json_file
from .....utils.logging import info, warning

if is_dep_available("tqdm"):
    from tqdm import tqdm


class Indexer(object):
    """Indexer"""

    def __init__(self):
        """init indexer"""
        self._map = {}
        self.idx = 0

    def get_id(self, key):
        """get id by key"""
        if key not in self._map:
            self.idx += 1
            self._map[key] = self.idx
        return self._map[key]

    def get_list(self, key_name):
        """return list containing key and id"""
        map_list = []
        for key in self._map:
            val = self._map[key]
            map_list.append({key_name: key, "id": val})
        return map_list


class Extension(object):
    """Extension"""

    def __init__(self, exts_list):
        """init extension"""
        self._exts_list = ["." + ext for ext in exts_list]

    def __iter__(self):
        """iterator"""
        return iter(self._exts_list)

    def update(self, ext):
        """update extension"""
        self._exts_list.remove(ext)
        self._exts_list.insert(0, ext)


def check_src_dataset(root_dir, dataset_type):
    """check src dataset format validity"""
    if dataset_type in ("VOC", "VOCWithUnlabeled"):
        pass
    elif dataset_type in ("LabelMe", "LabelMeWithUnlabeled"):
        pass
    else:
        raise ConvertFailedError(
            message=f"数据格式转换失败！不支持{dataset_type}格式数据集。当前仅支持 VOC、LabelMe 和 VOCWithUnlabeled、LabelMeWithUnlabeled 格式。"
        )

    err_msg_prefix = f"数据格式转换失败！请参考上述`{dataset_type}格式数据集示例`检查待转换数据集格式。"

    anno_map = {}
    for dst_anno, src_anno in [
        ("instance_train.json", "train_anno_list.txt"),
        ("instance_val.json", "val_anno_list.txt"),
    ]:
        src_anno_path = os.path.join(root_dir, src_anno)
        if not os.path.exists(src_anno_path):
            if dst_anno == "instance_train.json":
                raise ConvertFailedError(
                    message=f"{err_msg_prefix}保证{src_anno_path}文件存在。"
                )
            continue
        with custom_open(src_anno_path, "r") as f:
            anno_list = f.readlines()
        for anno_fn in anno_list:
            anno_fn = anno_fn.strip().split(" ")[-1]
            anno_path = os.path.join(root_dir, anno_fn)
            if not os.path.exists(anno_path):
                raise ConvertFailedError(
                    message=f'{err_msg_prefix}保证"{src_anno_path}"中的"{anno_fn}"文件存在。'
                )
        anno_map[dst_anno] = src_anno_path
    return anno_map


def convert(dataset_type, input_dir):
    """convert dataset to coco format"""
    # check format validity
    anno_map = check_src_dataset(input_dir, dataset_type)
    (
        convert_voc_dataset(input_dir, anno_map)
        if dataset_type in ("VOC", "VOCWithUnlabeled")
        else convert_labelme_dataset(input_dir, anno_map)
    )


def split_anno_list(root_dir, anno_map):
    """Split anno list to 80% train and 20% val"""

    train_anno_list = []
    val_anno_list = []
    anno_list_bak = os.path.join(root_dir, "train_anno_list.txt.bak")
    shutil.move(anno_map["instance_train.json"], anno_list_bak),
    with custom_open(anno_list_bak, "r") as f:
        src_anno = f.readlines()
    random.shuffle(src_anno)
    train_anno_list = src_anno[: int(len(src_anno) * 0.8)]
    val_anno_list = src_anno[int(len(src_anno) * 0.8) :]
    with custom_open(os.path.join(root_dir, "train_anno_list.txt"), "w") as f:
        f.writelines(train_anno_list)
    with custom_open(os.path.join(root_dir, "val_anno_list.txt"), "w") as f:
        f.writelines(val_anno_list)
    anno_map["instance_train.json"] = os.path.join(root_dir, "train_anno_list.txt")
    anno_map["instance_val.json"] = os.path.join(root_dir, "val_anno_list.txt")
    msg = f"{os.path.join(root_dir,'val_anno_list.txt')}不存在，数据集已默认按照80%训练集，20%验证集划分,\
        且将原始'train_anno_list.txt'重命名为'train_anno_list.txt.bak'."

    warning(msg)
    return anno_map


def convert_labelme_dataset(root_dir, anno_map):
    """convert dataset labeled by LabelMe to coco format"""
    label_indexer = Indexer()
    img_indexer = Indexer()

    annotations_dir = os.path.join(root_dir, "annotations")
    if not os.path.exists(annotations_dir):
        os.makedirs(annotations_dir)

    # FIXME(gaotingquan): support lmssl
    unlabeled_path = os.path.join(root_dir, "unlabeled.txt")
    if os.path.exists(unlabeled_path):
        shutil.move(unlabeled_path, os.path.join(annotations_dir, "unlabeled.txt"))

    # 不存在val_anno_list，对原始数据集进行划分
    if "instance_val.json" not in anno_map:
        anno_map = split_anno_list(root_dir, anno_map)

    for dst_anno in anno_map:
        labelme2coco(
            label_indexer,
            img_indexer,
            root_dir,
            anno_map[dst_anno],
            os.path.join(annotations_dir, dst_anno),
        )


@function_requires_deps("tqdm")
def labelme2coco(label_indexer, img_indexer, root_dir, anno_path, save_path):
    """convert json files generated by LabelMe to coco format and save to files"""
    with custom_open(anno_path, "r") as f:
        json_list = f.readlines()

    anno_num = 0
    anno_list = []
    image_list = []
    info(f"Start loading json annotation files from {anno_path} ...")
    for json_path in tqdm(json_list):
        json_path = json_path.strip()
        if not json_path.endswith(".json"):
            info(f'An illegal json path("{json_path}") found! Has been ignored.')
            continue

        with custom_open(os.path.join(root_dir, json_path.strip()), "r") as f:
            labelme_data = json.load(f)

        img_id = img_indexer.get_id(labelme_data["imagePath"])
        image_list.append(
            {
                "id": img_id,
                "file_name": labelme_data["imagePath"].split("/")[-1],
                "width": labelme_data["imageWidth"],
                "height": labelme_data["imageHeight"],
            }
        )

        for shape in labelme_data["shapes"]:
            assert shape["shape_type"] == "rectangle", "Only rectangle are supported."
            category_id = label_indexer.get_id(shape["label"])
            (x1, y1), (x2, y2) = shape["points"]
            x1, x2 = sorted([x1, x2])
            y1, y2 = sorted([y1, y2])
            bbox = list(map(float, [x1, y1, x2 - x1, y2 - y1]))
            anno_num += 1
            anno_list.append(
                {
                    "image_id": img_id,
                    "bbox": bbox,
                    "category_id": category_id,
                    "id": anno_num,
                    "iscrowd": 0,
                    "area": bbox[2] * bbox[3],
                    "ignore": 0,
                }
            )

    category_list = label_indexer.get_list(key_name="name")
    data_coco = {
        "images": image_list,
        "categories": category_list,
        "annotations": anno_list,
    }

    write_json_file(data_coco, save_path)
    info(f"The converted annotations has been save to {save_path}.")


def convert_voc_dataset(root_dir, anno_map):
    """convert VOC format dataset to coco format"""
    label_indexer = Indexer()
    img_indexer = Indexer()

    annotations_dir = os.path.join(root_dir, "annotations")
    if not os.path.exists(annotations_dir):
        os.makedirs(annotations_dir)

    # FIXME(gaotingquan): support lmssl
    unlabeled_path = os.path.join(root_dir, "unlabeled.txt")
    if os.path.exists(unlabeled_path):
        shutil.move(unlabeled_path, os.path.join(annotations_dir, "unlabeled.txt"))

    # 不存在val_anno_list，对原始数据集进行划分
    if "instance_val.json" not in anno_map:
        anno_map = split_anno_list(root_dir, anno_map)

    for dst_anno in anno_map:
        ann_paths = voc_get_label_anno(root_dir, anno_map[dst_anno])

        voc_xmls_to_cocojson(
            root_dir=root_dir,
            annotation_paths=ann_paths,
            label_indexer=label_indexer,
            img_indexer=img_indexer,
            output=annotations_dir,
            output_file=dst_anno,
        )


def voc_get_label_anno(root_dir, anno_path):
    """
    Read VOC format annotation file.

    Args:
        root_dir (str): The directory of VOC annotation file.
        anno_path (str): The annotation file path.

    Returns:
        tuple: A tuple of two elements, the first of which is of type dict, representing the mapping between tag names
        and their corresponding ids, and the second of type list, representing the list of paths to all annotated files.
    """
    if not os.path.exists(anno_path):
        info(f"The annotation file {anno_path} don't exists, has been ignored!")
        return []
    with custom_open(anno_path, "r") as f:
        ann_ids = f.readlines()

    ann_paths = []
    info(f"Start loading xml annotation files from {anno_path} ...")
    for aid in ann_ids:
        aid = aid.strip().split(" ")[-1]
        if not aid.endswith(".xml"):
            info(f'An illegal xml path("{aid}") found! Has been ignored.')
            continue
        ann_path = os.path.join(root_dir, aid)
        ann_paths.append(ann_path)

    return ann_paths


def voc_get_image_info(annotation_root, img_indexer):
    """
    Get the image info from VOC annotation file.

    Args:
        annotation_root: The annotation root.
        img_indexer: indexer to get image id by filename.

    Returns:
        dict: The image info.

    Raises:
        AssertionError: When filename cannot be found in 'annotation_root'.
    """
    filename = annotation_root.findtext("filename")
    assert filename is not None, filename
    os.path.basename(filename)
    im_id = img_indexer.get_id(filename)

    size = annotation_root.find("size")
    width = float(size.findtext("width"))
    height = float(size.findtext("height"))

    image_info = {"file_name": filename, "height": height, "width": width, "id": im_id}
    return image_info


def voc_get_coco_annotation(obj, label_indexer):
    """
    Convert VOC format annotation to COCO format.

    Args:
        obj: a obj in VOC.
        label_indexer: indexer to get category id by label name.

    Returns:
        dict: A dict with the COCO format annotation info.

    Raises:
        AssertionError: When the width or height of the annotation box is illegal.
    """
    label = obj.findtext("name")
    category_id = label_indexer.get_id(label)
    bndbox = obj.find("bndbox")
    xmin = float(bndbox.findtext("xmin"))
    ymin = float(bndbox.findtext("ymin"))
    xmax = float(bndbox.findtext("xmax"))
    ymax = float(bndbox.findtext("ymax"))
    if xmin > xmax or ymin > ymax:
        temp = xmin
        xmin = min(xmin, xmax)
        xmax = max(temp, xmax)
        temp = ymin
        ymin = min(ymin, ymax)
        ymax = max(temp, ymax)

    o_width = xmax - xmin
    o_height = ymax - ymin
    anno = {
        "area": o_width * o_height,
        "iscrowd": 0,
        "bbox": [xmin, ymin, o_width, o_height],
        "category_id": category_id,
        "ignore": 0,
    }
    return anno


@function_requires_deps("tqdm")
def voc_xmls_to_cocojson(
    root_dir, annotation_paths, label_indexer, img_indexer, output, output_file
):
    """
    Convert VOC format data to COCO format.

    Args:
        annotation_paths (list): A list of paths to the XML files.
        label_indexer: indexer to get category id by label name.
        img_indexer: indexer to get image id by filename.
        output (str): The directory to save output JSON file.
        output_file (str): Output JSON file name.

    Returns:
        None
    """
    extension_list = ["jpg", "png", "jpeg", "JPG", "PNG", "JPEG"]
    suffixs = Extension(extension_list)

    def match(root_dir, prefilename, prexlm_name):
        """matching extension"""
        for ext in suffixs:
            if os.path.exists(os.path.join(root_dir, "images", prefilename + ext)):
                suffixs.update(ext)
                return prefilename + ext
            elif os.path.exists(os.path.join(root_dir, "images", prexlm_name + ext)):
                suffixs.update(ext)
                return prexlm_name + ext
        return None

    output_json_dict = {
        "images": [],
        "type": "instances",
        "annotations": [],
        "categories": [],
    }

    bnd_id = 1  # bounding box start id
    info("Start converting !")
    for a_path in tqdm(annotation_paths):
        # Read annotation xml
        ann_tree = ET.parse(a_path)
        ann_root = ann_tree.getroot()
        file_name = ann_root.find("filename")
        prefile_name = file_name.text.split(".")[0]
        prexlm_name = os.path.basename(a_path).split(".")[0]
        # 根据file_name 和 xlm_name 分别匹配查找图片
        f_name = match(root_dir, prefile_name, prexlm_name)
        if f_name is not None:
            file_name.text = f_name
        else:
            prefile_name_set = set({prefile_name, prexlm_name})
            prefile_name_set = ",".join(prefile_name_set)
            suffix_set = ",".join(extension_list)
            images_path = os.path.join(root_dir, "images")
            info(
                f"{images_path}/{{{prefile_name_set}}}.{{{suffix_set}}} both not exists,will be skipped."
            )
            continue
        img_info = voc_get_image_info(ann_root, img_indexer)
        output_json_dict["images"].append(img_info)

        for obj in ann_root.findall("object"):
            if obj.find("bndbox") is None:  # Skip the object without bndbox
                continue
            ann = voc_get_coco_annotation(obj=obj, label_indexer=label_indexer)
            ann.update({"image_id": img_info["id"], "id": bnd_id})
            output_json_dict["annotations"].append(ann)
            bnd_id = bnd_id + 1

    output_json_dict["categories"] = label_indexer.get_list(key_name="name")
    output_file = os.path.join(output, output_file)
    write_json_file(output_json_dict, output_file)
    info(f"The converted annotations has been save to {output_file}.")



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import numpy as np
import pandas as pd

from .....utils.deps import class_requires_deps, is_dep_available
from ....utils.benchmark import benchmark
from .funcs import load_from_dataframe, time_feature

if is_dep_available("joblib"):
    import joblib

__all__ = [
    "BuildTSDataset",
    "TSCutOff",
    "TSNormalize",
    "TimeFeature",
    "TStoArray",
    "TStoBatch",
]


@benchmark.timeit
class TSCutOff:
    """Truncates time series data to a specified length for training.

    This class provides a method to truncate or cut off time series data
    to a specified input length, optionally skipping some initial data
    points. This is useful for preparing data for training models that
    require a fixed input size.
    """

    def __init__(self, size: Dict[str, int]):
        """Initializes the TSCutOff with size configurations.

        Args:
            size (Dict[str, int]): Dictionary containing size configurations,
                including 'in_chunk_len' for the input chunk length and
                optionally 'skip_chunk_len' for the number of initial data
                points to skip.
        """
        super().__init__()
        self.size = size

    def __call__(self, ts_list: List) -> List:
        """Applies the cut off operation to a list of time series.

        Args:
            ts_list (List): List of time series data frames to be truncated.

        Returns:
            List: List of truncated time series data frames.
        """
        return [self.cutoff(ts) for ts in ts_list]

    def cutoff(self, ts: Any) -> Any:
        """Truncates a single time series data frame to the specified length.

        This method truncates the time series data to the specified input
        chunk length, optionally skipping some initial data points. It raises
        a ValueError if the time series is too short.

        Args:
            ts: A single time series data frame to be truncated.

        Returns:
            Any: The truncated time series data frame.

        Raises:
            ValueError: If the time series length is less than the required
            minimum length (input chunk length plus any skip chunk length).
        """
        skip_len = self.size.get("skip_chunk_len", 0)
        if len(ts) < self.size["in_chunk_len"] + skip_len:
            raise ValueError(
                f"The length of the input data is {len(ts)}, but it should be at least {self.size['in_chunk_len'] + self.size['skip_chunk_len']} for training."
            )
        ts_data = ts[-(self.size["in_chunk_len"] + skip_len) :]
        return ts_data


@benchmark.timeit
@class_requires_deps("joblib")
class TSNormalize:
    """Normalizes time series data using a pre-fitted scaler.

    This class normalizes specified columns of time series data using a
    pre-fitted scaler loaded from a specified path. It supports normalization
    of both target and feature columns as specified in the parameters.
    """

    def __init__(self, scale_path: str, params_info: Dict[str, Any]):
        """Initializes the TSNormalize with a scaler and normalization parameters.

        Args:
            scale_path (str): Path to the pre-fitted scaler object file.
            params_info (Dict[str, Any]): Dictionary containing information
                about which columns to normalize, including 'target_cols'
                and 'feature_cols'.
        """
        super().__init__()
        self.scaler = joblib.load(scale_path)
        self.params_info = params_info

    def __call__(self, ts_list: List[pd.DataFrame]) -> List[pd.DataFrame]:
        """Applies normalization to a list of time series data frames.

        Args:
            ts_list (List[pd.DataFrame]): List of time series data frames to be normalized.

        Returns:
            List[pd.DataFrame]: List of normalized time series data frames.
        """
        return [self.tsnorm(ts) for ts in ts_list]

    def tsnorm(self, ts: pd.DataFrame) -> pd.DataFrame:
        """Normalizes specified columns of a single time series data frame.

        This method applies the scaler to normalize the specified target
        and feature columns of the time series.

        Args:
            ts (pd.DataFrame): A single time series data frame to be normalized.

        Returns:
            pd.DataFrame: The normalized time series data frame.
        """
        if self.params_info.get("target_cols", None) is not None:
            ts[self.params_info["target_cols"]] = self.scaler.transform(
                ts[self.params_info["target_cols"]]
            )
        if self.params_info.get("feature_cols", None) is not None:
            ts[self.params_info["feature_cols"]] = self.scaler.transform(
                ts[self.params_info["feature_cols"]]
            )
        return ts


@benchmark.timeit
class BuildTSDataset:
    """Constructs a time series dataset from a list of time series data frames."""

    def __init__(self, params_info: Dict[str, Any]):
        """Initializes the BuildTSDataset with parameters for dataset construction.

        Args:
            params_info (Dict[str, Any]): Dictionary containing parameters for
                constructing the time series dataset.
        """
        super().__init__()
        self.params_info = params_info

    def __call__(self, ts_list: List) -> List:
        """Applies the dataset construction to a list of time series.

        Args:
            ts_list (List): List of time series data frames.

        Returns:
            List: List of constructed time series datasets.
        """
        return [self.buildtsdata(ts) for ts in ts_list]

    def buildtsdata(self, ts) -> Any:
        """Builds a time series dataset from a single time series data frame.

        Args:
            ts: A single time series data frame.

        Returns:
            Any: A constructed time series dataset.
        """
        ts_data = load_from_dataframe(ts, **self.params_info)
        return ts_data


@benchmark.timeit
class TimeFeature:
    """Extracts time features from time series data for forecasting."""

    def __init__(
        self, params_info: Dict[str, Any], size: Dict[str, int], holiday: bool = False
    ):
        """Initializes the TimeFeature extractor.

        Args:
            params_info (Dict[str, Any]): Dictionary containing frequency information.
            size (Dict[str, int]): Dictionary containing the output chunk length.
            holiday (bool, optional): Whether to include holiday features. Defaults to False.
        """
        super().__init__()
        self.freq = params_info["freq"]
        self.size = size
        self.holiday = holiday

    def __call__(self, ts_list: List) -> List:
        """Applies time feature extraction to a list of time series.

        Args:
            ts_list (List): List of time series data frames.

        Returns:
            List: List of time series with extracted time features.
        """
        return [self.timefeat(ts) for ts in ts_list]

    def timefeat(self, ts: Dict[str, Any]) -> Any:
        """Extracts time features from a single time series data frame.

        Args:
            ts: A single time series data frame.

        Returns:
            Any: The time series with added time features.
        """
        if not self.holiday:
            ts = time_feature(
                ts,
                self.freq,
                ["hourofday", "dayofmonth", "dayofweek", "dayofyear"],
                self.size["out_chunk_len"],
            )
        else:
            ts = time_feature(
                ts,
                self.freq,
                [
                    "minuteofhour",
                    "hourofday",
                    "dayofmonth",
                    "dayofweek",
                    "dayofyear",
                    "monthofyear",
                    "weekofyear",
                    "holidays",
                ],
                self.size["out_chunk_len"],
            )
        return ts


@benchmark.timeit
class TStoArray:
    """Converts time series data into arrays for model input."""

    def __init__(self, input_data: Dict[str, Any]):
        """Initializes the TStoArray converter.

        Args:
            input_data (Dict[str, Any]): Dictionary specifying the input data format.
        """
        super().__init__()
        self.input_data = input_data

    def __call__(self, ts_list: List[Dict[str, Any]]) -> List[List[np.ndarray]]:
        """Converts a list of time series data frames into arrays.

        Args:
            ts_list (List[Dict[str, Any]]): List of time series data frames.

        Returns:
            List[List[np.ndarray]]: List of lists of arrays for each time series.
        """
        return [self.tstoarray(ts) for ts in ts_list]

    def tstoarray(self, ts: Dict[str, Any]) -> List[np.ndarray]:
        """Converts a single time series data frame into arrays.

        Args:
            ts (Dict[str, Any]): A single time series data frame.

        Returns:
            List[np.ndarray]: List of arrays representing the time series data.
        """
        ts_list = []
        input_name = list(self.input_data.keys())
        input_name.sort()
        for key in input_name:
            ts_list.append(np.array(ts[key]).astype("float32"))

        return ts_list


@benchmark.timeit
class TStoBatch:
    """Convert a list of time series into batches for processing.

    This class provides a method to convert a list of time series data into
    batches. Each time series in the list is assumed to be a sequence of
    equal-length arrays or DataFrames.
    """

    def __call__(self, ts_list: List[np.ndarray]) -> List[np.ndarray]:
        """Convert a list of time series into batches.

        This method stacks time series data along a new axis to create batches.
        It assumes that each time series in the list has the same length.

        Args:
            ts_list (List[np.ndarray]): A list of time series, where each
                time series is represented as a list or array of equal length.

        Returns:
            List[np.ndarray]: A list of batches, where each batch is a stacked
            array of time series data at the same index across all series.
        """
        n = len(ts_list[0])
        return [np.stack([ts[i] for ts in ts_list], axis=0) for i in range(n)]



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import numpy as np

from ..components import convert_points_to_boxes
from ..layout_parsing.utils import get_sub_regions_ocr_res
from ..ocr.result import OCRResult
from .result import SingleTableRecognitionResult


def get_ori_image_coordinate(x: int, y: int, box_list: list) -> list:
    """
    get the original coordinate from Cropped image to Original image.
    Args:
        x (int): x coordinate of cropped image
        y (int): y coordinate of cropped image
        box_list (list): list of table bounding boxes, eg. [[x1, y1, x2, y2, x3, y3, x4, y4]]
    Returns:
        list: list of original coordinates, eg. [[x1, y1, x2, y2, x3, y3, x4, y4]]
    """
    if not box_list:
        return box_list
    offset = np.array([x, y] * 4)
    box_list = np.array(box_list)
    if box_list.shape[-1] == 2:
        offset = offset.reshape(4, 2)
    ori_box_list = offset + box_list
    return ori_box_list


def convert_table_structure_pred_bbox(
    cell_points_list: list, crop_start_point: list, img_shape: tuple
) -> None:
    """
    Convert the predicted table structure bounding boxes to the original image coordinate system.

    Args:
        cell_points_list (list):  Bounding boxes ('bbox').
        crop_start_point (list): A list of two integers representing the starting point (x, y) of the cropped image region.
        img_shape (tuple): A tuple of two integers representing the shape (height, width) of the original image.

    Returns:
        cell_points_list (list):  Bounding boxes ('bbox').
    """

    ori_cell_points_list = get_ori_image_coordinate(
        crop_start_point[0], crop_start_point[1], cell_points_list
    )
    ori_cell_points_list = np.reshape(ori_cell_points_list, (-1, 4, 2))
    cell_box_list = convert_points_to_boxes(ori_cell_points_list)

    img_height, img_width = img_shape
    cell_box_list = np.clip(
        cell_box_list, 0, [img_width, img_height, img_width, img_height]
    )
    return cell_box_list


def distance(box_1: list, box_2: list) -> float:
    """
    compute the distance between two boxes

    Args:
        box_1 (list): first rectangle box,eg.(x1, y1, x2, y2)
        box_2 (list): second rectangle box,eg.(x1, y1, x2, y2)

    Returns:
        float: the distance between two boxes
    """
    x1, y1, x2, y2 = box_1
    x3, y3, x4, y4 = box_2
    center1_x = (x1 + x2) / 2
    center1_y = (y1 + y2) / 2
    center2_x = (x3 + x4) / 2
    center2_y = (y3 + y4) / 2
    dis = math.sqrt((center2_x - center1_x) ** 2 + (center2_y - center1_y) ** 2)
    dis_2 = abs(x3 - x1) + abs(y3 - y1)
    dis_3 = abs(x4 - x2) + abs(y4 - y2)
    return dis + min(dis_2, dis_3)


def compute_iou(rec1: list, rec2: list) -> float:
    """
    computing IoU
    Args:
        rec1 (list): (x1, y1, x2, y2)
        rec2 (list): (x1, y1, x2, y2)
    Returns:
        float: Intersection over Union
    """
    # computing area of each rectangles
    S_rec1 = (rec1[2] - rec1[0]) * (rec1[3] - rec1[1])
    S_rec2 = (rec2[2] - rec2[0]) * (rec2[3] - rec2[1])

    # computing the sum_area
    sum_area = S_rec1 + S_rec2

    # find the each edge of intersect rectangle
    left_line = max(rec1[0], rec2[0])
    right_line = min(rec1[2], rec2[2])
    top_line = max(rec1[1], rec2[1])
    bottom_line = min(rec1[3], rec2[3])

    # judge if there is an intersect
    if left_line >= right_line or top_line >= bottom_line:
        return 0.0
    else:
        intersect = (right_line - left_line) * (bottom_line - top_line)
        return (intersect / (sum_area - intersect)) * 1.0


def compute_inter(rec1, rec2):
    """
    computing intersection over rec2_area
    Args:
        rec1 (list): (x1, y1, x2, y2)
        rec2 (list): (x1, y1, x2, y2)
    Returns:
        float: Intersection over rec2_area
    """
    x1_1, y1_1, x2_1, y2_1 = map(float, rec1)
    x1_2, y1_2, x2_2, y2_2 = map(float, rec2)
    x_left = max(x1_1, x1_2)
    y_top = max(y1_1, y1_2)
    x_right = min(x2_1, x2_2)
    y_bottom = min(y2_1, y2_2)
    inter_width = max(0, x_right - x_left)
    inter_height = max(0, y_bottom - y_top)
    inter_area = inter_width * inter_height
    rec2_area = (x2_2 - x1_2) * (y2_2 - y1_2)
    if rec2_area == 0:
        return 0
    iou = inter_area / rec2_area
    return iou


def match_table_and_ocr(cell_box_list, ocr_dt_boxes, table_cells_flag, row_start_index):
    """
    match table and ocr

    Args:
        cell_box_list (list): bbox for table cell, 2 points, [left, top, right, bottom]
        ocr_dt_boxes (list): bbox for ocr, 2 points, [left, top, right, bottom]

    Returns:
        dict: matched dict, key is table index, value is ocr index
    """
    all_matched = []
    for k in range(len(table_cells_flag) - 1):
        matched = {}
        for i, table_box in enumerate(
            cell_box_list[table_cells_flag[k] : table_cells_flag[k + 1]]
        ):
            if len(table_box) == 8:
                table_box = [
                    np.min(table_box[0::2]),
                    np.min(table_box[1::2]),
                    np.max(table_box[0::2]),
                    np.max(table_box[1::2]),
                ]
            for j, ocr_box in enumerate(np.array(ocr_dt_boxes)):
                if compute_inter(table_box, ocr_box) > 0.7:
                    if i not in matched.keys():
                        matched[i] = [j]
                    else:
                        matched[i].append(j)
        real_len = max(matched.keys()) + 1 if len(matched) != 0 else 0
        if table_cells_flag[k + 1] < row_start_index[k + 1]:
            for s in range(row_start_index[k + 1] - table_cells_flag[k + 1]):
                matched[real_len + s] = []
        elif table_cells_flag[k + 1] > row_start_index[k + 1]:
            for s in range(table_cells_flag[k + 1] - row_start_index[k + 1]):
                matched[real_len - 1].append(matched[real_len + s])
        all_matched.append(matched)
    return all_matched


def get_html_result(
    all_matched_index: dict, ocr_contents: dict, pred_structures: list, table_cells_flag
) -> str:
    """
    Generates HTML content based on the matched index, OCR contents, and predicted structures.

    Args:
        matched_index (dict): A dictionary containing matched indices.
        ocr_contents (dict): A dictionary of OCR contents.
        pred_structures (list): A list of predicted HTML structures.

    Returns:
        str: Generated HTML content as a string.
    """
    pred_html = []
    td_index = 0
    td_count = 0
    matched_list_index = 0
    head_structure = pred_structures[0:3]
    html = "".join(head_structure)
    table_structure = pred_structures[3:-3]
    for tag in table_structure:
        matched_index = all_matched_index[matched_list_index]
        if "</td>" in tag:
            if "<td></td>" == tag:
                pred_html.extend("<td>")
            if td_index in matched_index.keys():
                if len(matched_index[td_index]) == 0:
                    continue
                b_with = False
                if (
                    "<b>" in ocr_contents[matched_index[td_index][0]]
                    and len(matched_index[td_index]) > 1
                ):
                    b_with = True
                    pred_html.extend("<b>")
                for i, td_index_index in enumerate(matched_index[td_index]):
                    content = ocr_contents[td_index_index]
                    if len(matched_index[td_index]) > 1:
                        if len(content) == 0:
                            continue
                        if content[0] == " ":
                            content = content[1:]
                        if "<b>" in content:
                            content = content[3:]
                        if "</b>" in content:
                            content = content[:-4]
                        if len(content) == 0:
                            continue
                        if i != len(matched_index[td_index]) - 1 and " " != content[-1]:
                            content += " "
                    pred_html.extend(content)
                if b_with:
                    pred_html.extend("</b>")
            if "<td></td>" == tag:
                pred_html.append("</td>")
            else:
                pred_html.append(tag)
            td_index += 1
            td_count += 1
            if (
                td_count >= table_cells_flag[matched_list_index + 1]
                and matched_list_index < len(all_matched_index) - 1
            ):
                matched_list_index += 1
                td_index = 0
        else:
            pred_html.append(tag)
    html += "".join(pred_html)
    end_structure = pred_structures[-3:]
    html += "".join(end_structure)
    return html


def sort_table_cells_boxes(boxes):
    """
    Sort the input list of bounding boxes.

    Args:
        boxes (list of lists): The input list of bounding boxes, where each bounding box is formatted as [x1, y1, x2, y2].

    Returns:
        sorted_boxes (list of lists): The list of bounding boxes sorted.
    """

    boxes_sorted_by_y = sorted(boxes, key=lambda box: box[1])
    rows = []
    current_row = []
    current_y = None
    tolerance = 10
    for box in boxes_sorted_by_y:
        x1, y1, x2, y2 = box
        if current_y is None:
            current_row.append(box)
            current_y = y1
        else:
            if abs(y1 - current_y) <= tolerance:
                current_row.append(box)
            else:
                current_row.sort(key=lambda x: x[0])
                rows.append(current_row)
                current_row = [box]
                current_y = y1
    if current_row:
        current_row.sort(key=lambda x: x[0])
        rows.append(current_row)
    sorted_boxes = []
    flag = [0]
    for i in range(len(rows)):
        sorted_boxes.extend(rows[i])
        if i < len(rows):
            flag.append(flag[i] + len(rows[i]))
    return sorted_boxes, flag


def convert_to_four_point_coordinates(boxes):
    """
    Convert bounding boxes from [x1, y1, x2, y2] format to
    [x1, y1, x2, y1, x2, y2, x1, y2] format.

    Parameters:
    - boxes: A list of bounding boxes, each defined as a list of integers
             in the format [x1, y1, x2, y2].

    Returns:
    - A list of bounding boxes, each converted to the format
      [x1, y1, x2, y1, x2, y2, x1, y2].
    """
    # Initialize an empty list to store the converted bounding boxes
    converted_boxes = []

    # Loop over each box in the input list
    for box in boxes:
        x1, y1, x2, y2 = box

        # Define the four corner points
        top_left = (x1, y1)
        top_right = (x2, y1)
        bottom_right = (x2, y2)
        bottom_left = (x1, y2)

        # Create a new list for the converted box
        converted_box = [
            top_left[0],
            top_left[1],  # Top-left corner
            top_right[0],
            top_right[1],  # Top-right corner
            bottom_right[0],
            bottom_right[1],  # Bottom-right corner
            bottom_left[0],
            bottom_left[1],  # Bottom-left corner
        ]

        # Append the converted box to the list
        converted_boxes.append(converted_box)

    return converted_boxes


def find_row_start_index(html_list):
    """
    find the index of the first cell in each row

    Args:
        html_list (list): list for html results

    Returns:
        row_start_indices (list): list for the index of the first cell in each row
    """
    # Initialize an empty list to store the indices of row start positions
    row_start_indices = []
    # Variable to track the current index in the flattened HTML content
    current_index = 0
    # Flag to check if we are inside a table row
    inside_row = False
    # Iterate through the HTML tags
    for keyword in html_list:
        # If a new row starts, set the inside_row flag to True
        if keyword == "<tr>":
            inside_row = True
        # If we encounter a closing row tag, set the inside_row flag to False
        elif keyword == "</tr>":
            inside_row = False
        # If we encounter a cell and we are inside a row
        elif (keyword == "<td></td>" or keyword == "</td>") and inside_row:
            # Append the current index as the starting index of the row
            row_start_indices.append(current_index)
            # Set the flag to ensure we only record the first cell of the current row
            inside_row = False
        # Increment the current index if we encounter a cell regardless of being inside a row or not
        if keyword == "<td></td>" or keyword == "</td>":
            current_index += 1
    # Return the computed starting indices of each row
    return row_start_indices


def map_and_get_max(table_cells_flag, row_start_index):
    """
    Retrieve table recognition result from cropped image info, table structure prediction, and overall OCR result.

    Args:
        table_cells_flag (list): List of the flags representing the end of each row of the table cells detection results.
        row_start_index (list): List of the flags representing the end of each row of the table structure predicted results.

    Returns:
        max_values: List of the process results.
    """

    max_values = []
    i = 0
    max_value = None
    for j in range(len(row_start_index)):
        while i < len(table_cells_flag) and table_cells_flag[i] <= row_start_index[j]:
            if max_value is None or table_cells_flag[i] > max_value:
                max_value = table_cells_flag[i]
            i += 1
        max_values.append(max_value if max_value is not None else row_start_index[j])
    return max_values


def get_table_recognition_res(
    table_box: list,
    table_structure_result: list,
    table_cells_result: list,
    overall_ocr_res: OCRResult,
    table_ocr_pred: dict,
    cells_texts_list: list,
    use_table_cells_ocr_results: bool,
    use_table_cells_split_ocr: bool,
) -> SingleTableRecognitionResult:
    """
    Retrieve table recognition result from cropped image info, table structure prediction, and overall OCR result.

    Args:
        table_box (list): Information about the location of cropped image, including the bounding box.
        table_structure_result (list): Predicted table structure.
        table_cells_result (list): Predicted table cells.
        overall_ocr_res (OCRResult): Overall OCR result from the input image.
        table_ocr_pred (dict): Table OCR result from the input image.
        cells_texts_list (list): OCR results with cells.
        use_table_cells_ocr_results (bool): whether to use OCR results with cells.

    Returns:
        SingleTableRecognitionResult: An object containing the single table recognition result.
    """

    table_cells_result = convert_to_four_point_coordinates(table_cells_result)
    table_box = np.array([table_box])

    if not (use_table_cells_ocr_results == True and use_table_cells_split_ocr == True):
        table_ocr_pred = get_sub_regions_ocr_res(overall_ocr_res, table_box)

    crop_start_point = [table_box[0][0], table_box[0][1]]
    img_shape = overall_ocr_res["doc_preprocessor_res"]["output_img"].shape[0:2]

    if len(table_cells_result) == 0 or len(table_ocr_pred["rec_boxes"]) == 0:
        pred_html = " ".join(table_structure_result)
        if len(table_cells_result) != 0:
            table_cells_result = convert_table_structure_pred_bbox(
                table_cells_result, crop_start_point, img_shape
            )
        single_img_res = {
            "cell_box_list": table_cells_result,
            "table_ocr_pred": table_ocr_pred,
            "pred_html": pred_html,
        }
        return SingleTableRecognitionResult(single_img_res)

    table_cells_result = convert_table_structure_pred_bbox(
        table_cells_result, crop_start_point, img_shape
    )

    if use_table_cells_ocr_results == True and use_table_cells_split_ocr == False:
        ocr_dt_boxes = table_cells_result
        ocr_texts_res = cells_texts_list
    else:
        ocr_dt_boxes = table_ocr_pred["rec_boxes"]
        ocr_texts_res = table_ocr_pred["rec_texts"]

    table_cells_result, table_cells_flag = sort_table_cells_boxes(table_cells_result)
    row_start_index = find_row_start_index(table_structure_result)
    table_cells_flag = map_and_get_max(table_cells_flag, row_start_index)
    table_cells_flag.append(len(table_cells_result))
    row_start_index.append(len(table_cells_result))
    matched_index = match_table_and_ocr(
        table_cells_result, ocr_dt_boxes, table_cells_flag, table_cells_flag
    )
    pred_html = get_html_result(
        matched_index, ocr_texts_res, table_structure_result, row_start_index
    )

    single_img_res = {
        "cell_box_list": table_cells_result,
        "table_ocr_pred": table_ocr_pred,
        "pred_html": pred_html,
    }

    return SingleTableRecognitionResult(single_img_res)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from .....utils.deps import function_requires_deps, is_dep_available
from .....utils.errors import ConvertFailedError
from .....utils.logging import info, warning

if is_dep_available("pycocotools"):
    from pycocotools.coco import COCO
if is_dep_available("tqdm"):
    from tqdm import tqdm


def check_src_dataset(root_dir, dataset_type):
    """check src dataset format validity"""
    if dataset_type in ("COCO"):
        pass
    else:
        raise ConvertFailedError(
            message=f"数据格式转换失败！不支持{dataset_type}格式数据集。当前仅支持 COCO 格式。"
        )

    err_msg_prefix = f"数据格式转换失败！请参考上述`{dataset_type}格式数据集示例`检查待转换数据集格式。"

    for anno in ["annotations/instance_train.json", "annotations/instance_val.json"]:
        src_anno_path = os.path.join(root_dir, anno)
        if not os.path.exists(src_anno_path):
            raise ConvertFailedError(
                message=f"{err_msg_prefix}保证{src_anno_path}文件存在。"
            )
    return None


def convert(dataset_type, input_dir):
    """convert dataset to multilabel format"""
    # check format validity
    check_src_dataset(input_dir, dataset_type)

    if dataset_type in ("COCO"):
        convert_coco_dataset(input_dir)
    else:
        raise ConvertFailedError(
            message=f"数据格式转换失败！不支持{dataset_type}格式数据集。当前仅支持 COCO 格式。"
        )


def convert_coco_dataset(root_dir):
    for anno in ["annotations/instance_train.json", "annotations/instance_val.json"]:
        src_img_dir = root_dir
        src_anno_path = os.path.join(root_dir, anno)
        coco2multilabels(src_img_dir, src_anno_path, root_dir)


@function_requires_deps("pycocotools", "tqdm")
def coco2multilabels(src_img_dir, src_anno_path, root_dir):
    image_dir = os.path.join(root_dir, "images")
    label_type = (
        os.path.basename(src_anno_path).replace("instance_", "").replace(".json", "")
    )
    anno_save_path = os.path.join(root_dir, "{}.txt".format(label_type))
    coco = COCO(src_anno_path)
    cat_id_map = {
        old_cat_id: new_cat_id for new_cat_id, old_cat_id in enumerate(coco.getCatIds())
    }
    num_classes = len(list(cat_id_map.keys()))

    with open(anno_save_path, "w") as fp:
        for img_id in tqdm(sorted(coco.getImgIds())):
            img_info = coco.loadImgs([img_id])[0]
            img_filename = img_info["file_name"]
            img_w = img_info["width"]
            img_h = img_info["height"]

            img_filepath = os.path.join(image_dir, img_filename)
            if not os.path.exists(img_filepath):
                warning(
                    "Illegal image file: {}, "
                    "and it will be ignored".format(img_filepath)
                )
                continue

            if img_w < 0 or img_h < 0:
                warning(
                    "Illegal width: {} or height: {} in annotation, "
                    "and im_id: {} will be ignored".format(img_w, img_h, img_id)
                )
                continue

            ins_anno_ids = coco.getAnnIds(imgIds=[img_id])
            instances = coco.loadAnns(ins_anno_ids)

            label = [0] * num_classes
            for instance in instances:
                label[cat_id_map[instance["category_id"]]] = 1
            img_filename = os.path.join("images", img_filename)
            fp.writelines("{}\t{}\n".format(img_filename, ",".join(map(str, label))))
        fp.close()
    if label_type == "train":
        label_txt_save_path = os.path.join(root_dir, "label.txt")
        with open(label_txt_save_path, "w") as fp:
            for cat in coco.cats.values():
                id = cat["id"]
                name = cat["name"]
                fp.writelines("{} {}\n".format(id, name))
            fp.close()
            info("Save label names to {}.".format(label_txt_save_path))



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import PIL
from PIL import ImageFont

from . import logging
from .cache import CACHE_DIR
from .download import download
from .flags import LOCAL_FONT_FILE_PATH


def create_font(txt: str, sz: tuple, font_path: str) -> ImageFont:
    """
    Create a font object with specified size and path, adjusted to fit within the given image region.

    Parameters:
    txt (str): The text to be rendered with the font.
    sz (tuple): A tuple containing the height and width of an image region, used for font size.
    font_path (str): The path to the font file.

    Returns:
    ImageFont: An ImageFont object adjusted to fit within the given image region.
    """

    font_size = int(sz[1] * 0.8)
    font = ImageFont.truetype(font_path, font_size, encoding="utf-8")
    if int(PIL.__version__.split(".")[0]) < 10:
        length = font.getsize(txt)[0]
    else:
        length = font.getlength(txt)

    if length > sz[0]:
        font_size = int(font_size * sz[0] / length)
        font = ImageFont.truetype(font_path, font_size, encoding="utf-8")
    return font


def create_font_vertical(
    txt: str, sz: tuple, font_path: str, scale=1.2
) -> ImageFont.FreeTypeFont:
    n = len(txt) if len(txt) > 0 else 1
    base_font_size = int(sz[1] / n * 0.8 * scale)
    base_font_size = max(base_font_size, 10)
    font = ImageFont.truetype(font_path, base_font_size, encoding="utf-8")

    if int(PIL.__version__.split(".")[0]) < 10:
        max_char_width = max([font.getsize(c)[0] for c in txt])
    else:
        max_char_width = max([font.getlength(c) for c in txt])

    if max_char_width > sz[0]:
        new_size = int(base_font_size * sz[0] / max_char_width)
        new_size = max(new_size, 10)
        font = ImageFont.truetype(font_path, new_size, encoding="utf-8")

    return font


class Font:
    def __init__(self, font_name=None, local_path=None):
        if local_path is None:
            if Path(str(LOCAL_FONT_FILE_PATH)).is_file():
                local_path = str(LOCAL_FONT_FILE_PATH)
        self._local_path = local_path
        if not local_path:
            assert font_name is not None
            self._font_name = font_name

    @property
    def path(self):
        # HACK: download font file when needed only
        if not self._local_path:
            self._get_offical_font()
        return self._local_path

    def _get_offical_font(self):
        """
        Download the official font file.
        """
        font_path = (Path(CACHE_DIR) / "fonts" / self._font_name).resolve().as_posix()
        if not Path(font_path).is_file():
            download(
                url=f"https://paddle-model-ecology.bj.bcebos.com/paddlex/PaddleX3.0/fonts/{self._font_name}",
                save_path=font_path,
            )
        self._local_path = font_path


if Path(str(LOCAL_FONT_FILE_PATH)).is_file():
    logging.warning(
        f"Using the local font file(`{LOCAL_FONT_FILE_PATH}`) specified by `LOCAL_FONT_FILE_PATH`!"
    )

PINGFANG_FONT = Font(font_name="PingFang-SC-Regular.ttf")
SIMFANG_FONT = Font(font_name="simfang.ttf")
LATIN_FONT = Font(font_name="latin.ttf")
TH_FONT = Font(font_name="th.ttf")
EL_FONT = Font(font_name="el.ttf")
KOREAN_FONT = Font(font_name="korean.ttf")
ARABIC_FONT = Font(font_name="arabic.ttf")
CYRILLIC_FONT = Font(font_name="cyrillic.ttf")
KANNADA_FONT = Font(font_name="kannada.ttf")
TELUGU_FONT = Font(font_name="telugu.ttf")
TAMIL_FONT = Font(font_name="tamil.ttf")
DEVANAGARI_FONT = Font(font_name="devanagari.ttf")



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import paddle
import paddle.nn.functional as F

try:
    from paddle.incubate.nn.functional import fused_rotary_position_embedding
except ImportError:
    fused_rotary_position_embedding = None

try:
    from paddle.incubate.nn.functional import swiglu
except ImportError:

    def swiglu(x, y=None):
        if y is None:
            x, y = paddle.chunk(x, chunks=2, axis=-1)
        return F.silu(x) * y


from paddle.utils import try_import


def get_env_device():
    """
    Return the device name of running environment.
    """
    if paddle.is_compiled_with_cuda():
        return "gpu"
    elif "npu" in paddle.device.get_all_custom_device_type():
        return "npu"
    elif "mlu" in paddle.device.get_all_custom_device_type():
        return "mlu"
    elif "gcu" in paddle.device.get_all_custom_device_type():
        return "gcu"
    elif "intel_hpu" in paddle.device.get_all_custom_device_type():
        return "intel_hpu"
    elif paddle.is_compiled_with_rocm():
        return "rocm"
    elif paddle.is_compiled_with_xpu():
        return "xpu"
    return "cpu"


try:
    from paddle.incubate.nn.functional import fused_rotary_position_embedding
except ImportError:
    fused_rotary_position_embedding = None
try:
    if get_env_device() in ["npu", "mlu", "gcu"]:
        from paddle.base import core

        for lib in os.listdir(os.getenv("CUSTOM_DEVICE_ROOT")):
            if lib.endswith(".so"):
                paddle.utils.cpp_extension.extension_utils.load_op_meta_info_and_register_op(
                    lib
                )
    from paddle.nn.functional.flash_attention import flash_attention
except:
    flash_attention = None


def fusion_rope(
    query_states,
    key_states,
    value_states,
    hidden_states,
    position_ids,
    past_key_value,
    rotary_emb,
    context_parallel_degree=-1,
):
    if get_env_device() not in ["gcu", "intel_hpu"]:
        assert past_key_value is None, "fuse rotary not support cache kv for now"
    batch_size, seq_length, num_heads, head_dim = query_states.shape
    _, kv_seq_len, num_key_value_heads, _ = key_states.shape
    if context_parallel_degree > 1:
        assert (
            get_env_device() == "gpu"
        ), "context parallel only support cuda device for now"
        kv_seq_len *= context_parallel_degree
    if get_env_device() not in ["gcu", "intel_hpu"]:
        cos, sin = rotary_emb(value_states, seq_len=kv_seq_len)
    if get_env_device() == "npu":
        query_states = core.eager._run_custom_op("fused_rope", query_states, cos, sin)[
            0
        ]
        key_states = core.eager._run_custom_op("fused_rope", key_states, cos, sin)[0]
    elif get_env_device() == "intel_hpu":
        if past_key_value is not None:
            kv_seq_len += past_key_value[0].shape[-3]
        cos, sin = rotary_emb(value_states, seq_len=kv_seq_len)
        cos = cos.squeeze().unsqueeze(0).unsqueeze(0)
        sin = sin.squeeze().unsqueeze(0).unsqueeze(0)
        query_states, _, _ = (
            paddle.incubate.nn.functional.fused_rotary_position_embedding(
                paddle.transpose(query_states, [0, 2, 1, 3]),
                None,
                None,
                sin=sin,
                cos=cos,
                position_ids=position_ids,
            )
        )
        key_states, _, _ = (
            paddle.incubate.nn.functional.fused_rotary_position_embedding(
                paddle.transpose(key_states, [0, 2, 1, 3]),
                None,
                None,
                sin=sin,
                cos=cos,
                position_ids=position_ids,
            )
        )
        query_states = paddle.transpose(query_states, [0, 2, 1, 3])
        key_states = paddle.transpose(key_states, [0, 2, 1, 3])
    elif get_env_device() == "gcu":
        cos_sin = rotary_emb.get_fused_cos_sin(value_states, seq_len=kv_seq_len)
        query_states, key_states = core.eager._run_custom_op(
            "fused_rotary_embedding_gcu",
            query_states,
            key_states,
            cos_sin,
            position_ids,
            True,
        )
    else:
        # paddle version > 2.6 or develop support q and k/v with different num_heads
        paddle_version = float(paddle.__version__[:3])
        if ((paddle_version != 0.0) and (paddle_version <= 2.6)) and (
            num_heads != num_key_value_heads
        ):
            query_states, _, _ = fused_rotary_position_embedding(
                query_states,
                None,
                None,
                sin=sin,
                cos=cos,
                position_ids=position_ids,
                use_neox_rotary_style=False,
            )
            key_states, _, _ = fused_rotary_position_embedding(
                key_states,
                None,
                None,
                sin=sin,
                cos=cos,
                position_ids=position_ids,
                use_neox_rotary_style=False,
            )
        else:
            query_states, key_states, _ = fused_rotary_position_embedding(
                query_states,
                key_states,
                v=None,
                sin=sin,
                cos=cos,
                position_ids=position_ids,
                use_neox_rotary_style=False,
            )
    return query_states, key_states


def rms_norm_fused(x_in, w, eps, use_fast_ln=False):
    if use_fast_ln:
        fast_ln = try_import("fast_ln")
        return fast_ln.fast_rms_norm(x_in, w, eps)[0]
    else:
        fused_ln = try_import("fused_ln")
        return fused_ln.fused_rms_norm(x_in, w, eps)[0]


def fusion_rms_norm(hidden_states, weight, variance_epsilon, use_fast_ln=False):
    if get_env_device() == "npu":
        return core.eager._run_custom_op(
            "rms_norm_npu", hidden_states, weight, variance_epsilon
        )[0]
    if get_env_device() == "mlu":
        return core.eager._run_custom_op(
            "rms_norm_mlu", hidden_states, weight, variance_epsilon
        )[0]
    elif get_env_device() == "gcu":
        return core.eager._run_custom_op(
            "rms_norm_gcu", hidden_states, weight, variance_epsilon
        )[0]
    elif get_env_device() == "intel_hpu":
        return paddle.incubate.nn.functional.fused_rms_norm(
            hidden_states, weight, None, variance_epsilon, hidden_states.dim() - 1
        )[0]

    return rms_norm_fused(hidden_states, weight, variance_epsilon, use_fast_ln)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional, TypeVar

import numpy as np
import paddle
from numpy import ndarray, transpose

from ......utils import logging
from ..distributed import distributed_allgather, distributed_gather
from ..utils import device_guard, get_env_device

if TYPE_CHECKING:
    from .configuration_utils import PretrainedConfig

# the type hinting for pytorch model & layer & tensor
Module = TypeVar("Module")
PytorchTensor = TypeVar("PytorchTensor")


@dataclass
class StateDictNameMapping:
    """NameMapping of StateDict between two models"""

    source_name: str
    target_name: str = None

    action: Optional[str] = None  # the value can be: transpose, merge_last_two_dim
    index: Optional[int] = None

    slots: list[str] = None

    def __post_init__(self):
        self.target_name = self.target_name or self.source_name

    def should_transpose(self) -> bool:
        return self.action == "transpose"

    def should_merge_last_two_dim(self) -> bool:
        """check that whether merge last two dim"""
        return self.action == "merge_last_two_dim"

    def run(self, state_dict: dict[str, ndarray], name: str) -> ndarray:
        """run some custom operation on ndarray, eg: transpose, merge_last_two_dim

        Args:
            tensor (ndarray): the source of the tensor data

        Returns:
            ndarray: the final tensor
        """
        tensor = state_dict.pop(name)
        if callable(self.action):
            return self.action(tensor)
        if self.action == "transpose":
            return transpose(tensor, [1, 0])
        if self.action == "merge_last_two_dim":
            shape = tensor.shape
            assert len(shape) == 3
            return np.reshape(tensor, [shape[0], -1])
        if self.action == "split":
            assert (
                self.index is not None
            ), "when action is `split`, index field is required."
            # FIXME if the order of split starts from index=2, no tensor left.
            if self.index < 2:
                state_dict[name] = tensor
            # qkv is stored in same tensor, so it should be split into 3 arr
            tensors = np.split(tensor, 3, axis=-1)
            return tensors[self.index]

        return tensor

    def matched(self, text: str) -> bool:
        """check whether the layer_name match the current pattern

        Args:
            text (str): the name of layer

        Returns:
            bool: whether the
        """
        if text == self.source_name:
            return True

        if not self.slots:
            return False


class ConversionMixin:
    @classmethod
    def support_conversion(cls, config: PretrainedConfig) -> bool:
        """check whether the model support conversion"""
        try:
            # try to get the name-mapping info
            _ = cls._get_name_mappings(config)
        except NotImplementedError:
            return False
        finally:
            return True

    @classmethod
    def _get_name_mappings(cls, config: PretrainedConfig) -> List[StateDictNameMapping]:
        """get name mapping of PretrainedModel

        Args:
            config (PretrainedConfig): the configuration of name-mapping

        Raises:
            NotImplementedError:

        Returns:
            List[StateDictNameMapping]: the name-mappings of pretrained model
        """
        raise NotImplementedError

    @classmethod
    def get_tensor_parallel_convert_actions(
        cls,
        config: PretrainedConfig,
        loaded_state_dict_keys,
        is_split=True,
        ignore_error=False,
    ):
        name_action_mappings = cls._get_tensor_parallel_mappings(
            config, is_split=is_split
        )
        state_keys_map = cls._resolve_prefix_keys(
            name_action_mappings.keys(), loaded_state_dict_keys, ignore_error
        )
        for k, v in state_keys_map.items():
            name_action_mappings[v] = name_action_mappings.pop(k)
        return name_action_mappings

    @classmethod
    def convert_tensor_parallel(
        cls,
        weight_file: str,
        config: PretrainedConfig,
        state_dict=None,
        ignore_error=False,
    ) -> None:
        """the entry of converting config and converting model file

        Args:
            weight_file (str | None): the weight file path of `model_state.pdparams` file
            config (PretrainedConfig): the PretrainedConfig instance of model
        """

        name_action_mappings = cls._get_tensor_parallel_mappings(config)
        if state_dict is None:
            with device_guard("cpu"):
                state_dict = paddle.load(weight_file, return_numpy=False)
            logging.info(
                "Starting to convert original state_dict to tensor parallel state_dict."
            )

        state_keys_map = cls._resolve_prefix_keys(
            name_action_mappings.keys(), state_dict.keys(), ignore_error
        )

        for k, v in state_keys_map.items():
            name_action_mappings[v] = name_action_mappings.pop(k)

        for name, action in name_action_mappings.items():
            if name not in state_dict:
                if not ignore_error:
                    logging.warning(f"Key <{name}> not in the model state weight file.")
                continue
            tensor = state_dict.pop(name)
            new_tensor = action(tensor)
            with device_guard("cpu"):
                state_dict[name] = paddle.Tensor(new_tensor, zero_copy=True)

        return state_dict

    @classmethod
    def merge_tensor_parallel(cls, state_dict, config) -> None:
        """the entry of converting config and converting model file

        Args:
            input_dir (str | None): the input dir which contains `pytorch_model.bin` and `config.json` file
            config (PretrainedConfig): the PretrainedConfig instance of model
        """
        name_action_mappings = cls._get_tensor_parallel_mappings(config, is_split=False)
        state_keys_map = cls._resolve_prefix_keys(
            name_action_mappings.keys(), state_dict.keys()
        )

        for k, v in state_keys_map.items():
            name_action_mappings[v] = name_action_mappings.pop(k)

        state_dict_to_save = {}

        hcg = paddle.distributed.fleet.get_hybrid_communicate_group()
        mp_group = hcg.get_model_parallel_group()
        is_dst = paddle.distributed.get_rank(mp_group) == 0

        for key in state_dict.keys():
            tensor = state_dict[key]
            if key in name_action_mappings:
                if get_env_device() == "xpu":
                    ret = distributed_allgather(tensor, group=mp_group, offload=True)
                else:
                    ret = distributed_gather(tensor, group=mp_group, offload=True)
                action = name_action_mappings.pop(key)
                tensor = action(ret) if is_dst else None
            else:
                tensor = tensor.cpu().numpy() if is_dst else None

            # keep state dict use paddle.tensor
            if isinstance(tensor, np.ndarray):
                with device_guard("cpu"):
                    tensor = paddle.Tensor(tensor, zero_copy=True)

            state_dict_to_save[key] = tensor

        if len(name_action_mappings) > 0:
            for x in name_action_mappings.keys():
                logging.debug(
                    f"key <{x}> need to merge tensor parallel but we can't find in model state."
                )

        return state_dict_to_save

    @classmethod
    def _get_tensor_parallel_mappings(
        cls, config: PretrainedConfig, is_split=True
    ) -> List[StateDictNameMapping]:
        """get name mapping of PretrainedModel

        Args:
            config (PretrainedConfig): the configuration of name-mapping

        Raises:
            NotImplementedError:

        Returns:
            List[StateDictNameMapping]: the name-mappings for tensor_parallel
        """
        raise NotImplementedError

    @staticmethod
    def _resolve_prefix_keys(state_keys_base, state_keys_real, ignore_error=False):
        # state_keys_map base to real
        state_keys_map = {}

        # sorted by length，match from long to short for A.key B.key ...
        state_keys_base = sorted(state_keys_base, key=lambda x: len(x), reverse=True)
        state_keys_real = set(state_keys_real)

        for key in state_keys_base:
            for x in state_keys_real:
                if x.endswith(key):
                    state_keys_map[key] = x
                    break
            if key not in state_keys_map:
                if not ignore_error:
                    logging.debug(
                        f"tensor parallel conversion: could not find name {key} in loaded state dict!"
                    )
            else:
                state_keys_real.remove(state_keys_map[key])

        return state_keys_map

    @classmethod
    def convert_fuse_and_split(
        cls, config: PretrainedConfig, state_dict, tp_actions=None
    ):
        loaded_keys = state_dict.keys()
        # collect and convert fuse/split action
        fused_and_split_keys = []
        convert_with_same_keys = []
        fuse_actions, resume_keys = cls.get_fuse_or_split_param_convert_actions(
            config, loaded_keys, is_fuse=True
        )
        for keys, action in fuse_actions.items():
            if keys[-1] in keys[:-1]:
                assert len(keys) == 2, "only 2 keys can be converted with the same name"
                convert_with_same_keys.append(keys[-1])
            origin_states = [state_dict.pop(key) for key in keys[:-1]]
            state_dict[keys[-1]] = action(origin_states)
            fused_and_split_keys.append(keys[-1])
            logging.debug(f"Fusing parameter: {keys[:-1]} into {keys[-1]}")

        split_actions, _ = cls.get_fuse_or_split_param_convert_actions(
            config, loaded_keys, is_fuse=False
        )
        for keys, action in split_actions.items():
            if keys[-1] in keys[:-1]:
                assert len(keys) == 2, "only 2 keys can be converted with the same name"
                convert_with_same_keys.append(keys[-1])
            origin_state = state_dict.pop(keys[-1])
            split_states = action(origin_state)
            for key_idx, key in enumerate(keys[:-1]):
                state_dict[key] = split_states[key_idx]
                fused_and_split_keys.append(key)
            logging.debug(f"Splitting parameter: {keys[-1]} into {keys[:-1]}")

        if tp_actions is not None:
            for key in fused_and_split_keys:
                if key in convert_with_same_keys:
                    continue

                for name in tp_actions.keys():
                    if key.endswith(name):
                        with device_guard():
                            state_dict[key] = paddle.Tensor(
                                tp_actions[name](state_dict.pop(key)), zero_copy=True
                            )
                        break

        resume_state_dict = {k: state_dict[k] for k in resume_keys if k in state_dict}
        return state_dict, resume_state_dict

    @classmethod
    def get_fuse_or_split_param_convert_actions(
        cls,
        config: PretrainedConfig,
        loaded_state_dict_keys,
        is_fuse=True,
        ignore_error=False,
    ):
        name_action_mappings = cls._get_fuse_or_split_param_mappings(config, is_fuse)
        state_keys_map = cls._resolve_prefix_keys_for_fuse_and_split(
            name_action_mappings.keys(), loaded_state_dict_keys, ignore_error, is_fuse
        )
        for k, v in state_keys_map.items():
            name_action_mappings[v] = name_action_mappings.pop(k)

        filter_name_action = {}
        resume_keys = []
        if is_fuse:
            for k, v in name_action_mappings.items():
                cond = True
                if not all(item in loaded_state_dict_keys for item in k[:-1]):
                    # resume keys for next fuse
                    resume_keys += k[:-1]
                    cond = False
                if cond:
                    filter_name_action[k] = v
        else:
            for k, v in name_action_mappings.items():
                if k[-1] in loaded_state_dict_keys:
                    filter_name_action[k] = v

        return filter_name_action, resume_keys

    @classmethod
    def _get_fuse_or_split_param_mappings(
        cls, config: PretrainedConfig, is_fuse=True
    ) -> List[StateDictNameMapping]:
        """get fused parameter mapping of PretrainedModel

        Args:
            config (PretrainedConfig): the configuration of name-mapping

        Raises:
            NotImplementedError:

        Returns:
            List[StateDictNameMapping]: the name-mappings for tensor_parallel
        """

        return {}

    @staticmethod
    def _resolve_prefix_keys_for_fuse_and_split(
        state_keys_base, state_keys_real, ignore_error=False, is_fuse=True
    ):
        state_keys_map = {}

        for keys in state_keys_base:
            prefix = ""
            if is_fuse:
                for x in state_keys_real:
                    for base_key in keys[:-1]:
                        if x.endswith(base_key):
                            prefix = x.replace(base_key, "")
                            break
                    if prefix != "":
                        break
            else:
                base_key = keys[-1]
                for x in state_keys_real:
                    if x.endswith(base_key):
                        prefix = x.replace(base_key, "")
                        break

            new_keys = tuple([prefix + key for key in keys])
            state_keys_map[keys] = new_keys

        return state_keys_map



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ...base import BaseRunner
from ...base.utils.subprocess import CompletedProcess


class ClsRunner(BaseRunner):
    """Cls Runner"""

    def train(
        self,
        config_path: str,
        cli_args: list,
        device: str,
        ips: str,
        save_dir: str,
        do_eval=True,
    ) -> CompletedProcess:
        """train model

        Args:
            config_path (str): the config file path used to train.
            cli_args (list): the additional parameters.
            device (str): the training device.
            ips (str): the ip addresses of nodes when using distribution.
            save_dir (str): the directory path to save training output.
            do_eval (bool, optional): whether or not to evaluate model during training. Defaults to True.

        Returns:
            CompletedProcess: the result of training subprocess execution.
        """
        args, env = self.distributed(device, ips, log_dir=save_dir)
        cmd = [*args, "tools/train.py", "-c", config_path, *cli_args]
        cmd.extend(["-o", f"Global.eval_during_train={do_eval}"])
        return self.run_cmd(
            cmd,
            env=env,
            switch_wdir=True,
            echo=True,
            silent=False,
            capture_output=True,
            log_path=self._get_train_log_path(save_dir),
        )

    def evaluate(
        self, config_path: str, cli_args: list, device: str, ips: str
    ) -> CompletedProcess:
        """run model evaluating

        Args:
            config_path (str): the config file path used to evaluate.
            cli_args (list): the additional parameters.
            device (str): the evaluating device.
            ips (str): the ip addresses of nodes when using distribution.

        Returns:
            CompletedProcess: the result of evaluating subprocess execution.
        """
        args, env = self.distributed(device, ips)
        cmd = [*args, "tools/eval.py", "-c", config_path, *cli_args]
        cp = self.run_cmd(
            cmd, env=env, switch_wdir=True, echo=True, silent=False, capture_output=True
        )
        if cp.returncode == 0:
            metric_dict = _extract_eval_metrics(cp.stdout)
            cp.metrics = metric_dict
        return cp

    def predict(
        self, config_path: str, cli_args: list, device: str
    ) -> CompletedProcess:
        """run predicting using dynamic mode

        Args:
            config_path (str): the config file path used to predict.
            cli_args (list): the additional parameters.
            device (str): unused.

        Returns:
            CompletedProcess: the result of predicting subprocess execution.
        """
        # `device` unused
        cmd = [self.python, "tools/infer.py", "-c", config_path, *cli_args]
        return self.run_cmd(cmd, switch_wdir=True, echo=True, silent=False)

    def export(
        self, config_path: str, cli_args: list, device: str, save_dir: str = None
    ) -> CompletedProcess:
        """run exporting

        Args:
            config_path (str): the path of config file used to export.
            cli_args (list): the additional parameters.
            device (str): unused.
            save_dir (str, optional): the directory path to save exporting output. Defaults to None.

        Returns:
            CompletedProcess: the result of exporting subprocess execution.
        """
        # `device` unused

        cmd = [
            self.python,
            "tools/export_model.py",
            "-c",
            config_path,
            *cli_args,
            "-o",
            "Global.export_for_fd=True",
        ]

        cp = self.run_cmd(cmd, switch_wdir=True, echo=True, silent=False)
        return cp

    def infer(self, config_path: str, cli_args: list, device: str) -> CompletedProcess:
        """run predicting using inference model

        Args:
            config_path (str): the path of config file used to predict.
            cli_args (list): the additional parameters.
            device (str): unused.

        Returns:
            CompletedProcess: the result of inferring subprocess execution.
        """
        # `device` unused
        cmd = [self.python, "python/predict_cls.py", "-c", config_path, *cli_args]
        return self.run_cmd(cmd, switch_wdir="deploy", echo=True, silent=False)

    def compression(
        self,
        config_path: str,
        train_cli_args: list,
        export_cli_args: list,
        device: str,
        train_save_dir: str,
    ) -> CompletedProcess:
        """run compression model

        Args:
            config_path (str): the path of config file used to predict.
            train_cli_args (list): the additional training parameters.
            export_cli_args (list): the additional exporting parameters.
            device (str): the running device.
            train_save_dir (str): the directory path to save output.

        Returns:
            CompletedProcess: the result of compression subprocess execution.
        """
        # Step 1: Train model
        cp_train = self.train(config_path, train_cli_args, device, None, train_save_dir)

        # Step 2: Export model
        weight_path = os.path.join(train_save_dir, "best_model", "model")
        export_cli_args = [
            *export_cli_args,
            "-o",
            f"Global.pretrained_model={weight_path}",
        ]
        cp_export = self.export(config_path, export_cli_args, device)

        return cp_train, cp_export


def _extract_eval_metrics(stdout: str) -> dict:
    """extract evaluation metrics from training log

    Args:
        stdout (str): the training log

    Returns:
        dict: the training metric
    """
    import re

    _DP = r"[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?"
    patterns = [
        r"\[Eval\]\[Epoch 0\]\[Avg\].*top1: (_dp)".replace("_dp", _DP),
        r"\[Eval\]\[Epoch 0\]\[Avg\].*top1: (_dp), top5: (_dp)".replace("_dp", _DP),
        r"\[Eval\]\[Epoch 0\]\[Avg\].*recall1: (_dp), recall5: (_dp), mAP: (_dp)".replace(
            "_dp", _DP
        ),
        r"\[Eval\]\[Epoch 0\]\[Avg\].*MultiLabelMAP\(integral\): (_dp)".replace(
            "_dp", _DP
        ),
        r"\[Eval\]\[Epoch 0\]\[Avg\].*evalres:\ ma: (_dp)".replace("_dp", _DP),
    ]
    keys = [
        ["val.top1"],
        ["val.top1", "val.top5"],
        ["recall1", "recall5", "mAP"],
        ["MultiLabelMAP"],
        ["evalres: ma"],
    ]

    metric_dict = dict()
    for pattern, key in zip(patterns, keys):
        pattern = re.compile(pattern)
        for line in stdout.splitlines():
            match = pattern.search(line)
            if match:
                for k, v in zip(key, map(float, match.groups())):
                    metric_dict[k] = v
    return metric_dict



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import os.path as osp

import pandas as pd

from .....utils.errors import ConvertFailedError


def check_src_dataset(root_dir):
    """check src dataset format validity"""

    err_msg_prefix = f"Convert Failed! Only support '.xlsx/.xls' format files."

    for dst_anno, src_anno in [("train.xlsx", "train.xls"), ("val.xlsx", "val.xls")]:
        src_anno_path = os.path.join(root_dir, src_anno)
        dst_anno_path = os.path.join(root_dir, dst_anno)
        if not os.path.exists(src_anno_path) and not os.path.exists(dst_anno_path):
            if "train" in dst_anno:
                raise ConvertFailedError(
                    message=f"{err_msg_prefix} Ensure {src_anno_path} or {dst_anno_path} exist."
                )
            continue


def convert_excel_dataset(input_dir):
    """
    将excel标注的数据集转换为PaddleX需要的格式

    Args:
        input_dir (str): 输入的目录，包含多个json格式的Labelme标注文件

    Returns:
        str: 返回一个字符串表示转换的结果，“转换成功”表示转换没有问题。

    Raises:
        该函数目前没有特定的异常抛出。

    """
    # read excel file
    for dst_anno, src_anno in [("train.xlsx", "train.xls"), ("val.xlsx", "val.xls")]:
        src_anno_path = os.path.join(input_dir, src_anno)
        dst_anno_path = os.path.join(input_dir, dst_anno)

        if os.path.exists(src_anno_path):
            excel_file = pd.read_excel(src_anno_path)
            output_csv_dir = osp.join(input_dir, src_anno.replace(".xls", ".csv"))
            excel_file.to_csv(output_csv_dir, index=False)
        elif os.path.exists(dst_anno_path):
            excel_file = pd.read_excel(dst_anno_path)
            output_csv_dir = osp.join(input_dir, dst_anno.replace(".xlsx", ".csv"))
            excel_file.to_csv(output_csv_dir, index=False)


def convert(input_dir):
    """convert dataset to coco format"""
    # check format validity
    check_src_dataset(input_dir)
    convert_excel_dataset(input_dir)
    return input_dir



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from typing import Any, Dict

from .....utils.deps import function_requires_deps, is_dep_available
from ...infra.config import create_app_config

if is_dep_available("fastapi"):
    from fastapi import FastAPI


def _pipeline_name_to_mod_name(pipeline_name: str) -> str:
    if not pipeline_name:
        raise ValueError("Empty pipeline name")
    mod_name = pipeline_name.lower().replace("-", "_")
    if mod_name[0].isdigit():
        return "m_" + mod_name
    return mod_name


# XXX: A dynamic approach is used here for writing fewer lines of code, at the
# cost of sacrificing some benefits of type hints.
@function_requires_deps("fastapi")
def create_pipeline_app(pipeline: Any, pipeline_config: Dict[str, Any]) -> "FastAPI":
    pipeline_name = pipeline_config["pipeline_name"]
    mod_name = _pipeline_name_to_mod_name(pipeline_name)
    mod = importlib.import_module(f".{mod_name}", package=__package__)
    app_config = create_app_config(pipeline_config)
    app_creator = getattr(mod, "create_pipeline_app")
    app = app_creator(pipeline, app_config)
    return app



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
import json
from queue import Queue
from threading import Thread
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
)

from typing_extensions import ParamSpec, TypeGuard

from ....utils import logging
from ....utils.deps import class_requires_deps, function_requires_deps, is_dep_available
from ...pipelines import BasePipeline
from ..infra.config import AppConfig
from ..infra.models import AIStudioNoResultResponse
from ..infra.utils import call_async, generate_log_id

if is_dep_available("aiohttp"):
    import aiohttp
if is_dep_available("fastapi"):
    import fastapi
    from fastapi.encoders import jsonable_encoder
    from fastapi.exceptions import RequestValidationError
    from fastapi.responses import JSONResponse
if is_dep_available("starlette"):
    from starlette.exceptions import HTTPException

PipelineT = TypeVar("PipelineT", bound=BasePipeline)
P = ParamSpec("P")
R = TypeVar("R")


class _Error(TypedDict):
    error: str


def _is_error(obj: object) -> TypeGuard[_Error]:
    return (
        isinstance(obj, dict)
        and obj.keys() == {"error"}
        and isinstance(obj["error"], str)
    )


# XXX: Since typing info (e.g., the pipeline class) cannot be easily obtained
# without abstraction leaks, generic classes do not offer additional benefits
# for type hinting. However, I would stick with the current design, as it does
# not introduce runtime overhead at the moment and may prove useful in the
# future.
@class_requires_deps("fastapi")
class PipelineWrapper(Generic[PipelineT]):
    def __init__(self, pipeline: PipelineT) -> None:
        super().__init__()
        self._pipeline = pipeline
        # HACK: We work around a bug in Paddle Inference by performing all
        # inference in the same thread.
        self._queue = Queue()
        self._closed = False
        self._loop = asyncio.get_running_loop()
        self._thread = Thread(target=self._worker, daemon=False)
        self._thread.start()

    @property
    def pipeline(self) -> PipelineT:
        return self._pipeline

    async def infer(self, *args: Any, **kwargs: Any) -> List[Any]:
        def _infer(*args, **kwargs) -> List[Any]:
            output: list = []
            with contextlib.closing(self._pipeline.predict(*args, **kwargs)) as it:
                for item in it:
                    if _is_error(item):
                        raise fastapi.HTTPException(
                            status_code=500, detail=item["error"]
                        )
                    output.append(item)

            return output

        return await self.call(_infer, *args, **kwargs)

    async def call(self, func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        if self._closed:
            raise RuntimeError("`PipelineWrapper` has already been closed")
        fut = self._loop.create_future()
        self._queue.put((func, args, kwargs, fut))
        return await fut

    async def close(self):
        if not self._closed:
            self._queue.put(None)
            await call_async(self._thread.join)
            self._closed = True

    def _worker(self):
        while not self._closed:
            item = self._queue.get()
            if item is None:
                break
            func, args, kwargs, fut = item
            try:
                result = func(*args, **kwargs)
                self._loop.call_soon_threadsafe(fut.set_result, result)
            except Exception as e:
                self._loop.call_soon_threadsafe(fut.set_exception, e)
            finally:
                self._queue.task_done()


@class_requires_deps("aiohttp")
class AppContext(Generic[PipelineT]):
    def __init__(self, *, config: AppConfig) -> None:
        super().__init__()
        self._config = config
        self.extra: Dict[str, Any] = {}
        self._pipeline: Optional[PipelineWrapper[PipelineT]] = None
        self._aiohttp_session: Optional[aiohttp.ClientSession] = None

    @property
    def config(self) -> AppConfig:
        return self._config

    @property
    def pipeline(self) -> PipelineWrapper[PipelineT]:
        if not self._pipeline:
            raise AttributeError("`pipeline` has not been set.")
        return self._pipeline

    @pipeline.setter
    def pipeline(self, val: PipelineWrapper[PipelineT]) -> None:
        self._pipeline = val

    @property
    def aiohttp_session(self) -> "aiohttp.ClientSession":
        if not self._aiohttp_session:
            raise AttributeError("`aiohttp_session` has not been set.")
        return self._aiohttp_session

    @aiohttp_session.setter
    def aiohttp_session(self, val: "aiohttp.ClientSession") -> None:
        self._aiohttp_session = val


@function_requires_deps("fastapi", "aiohttp", "starlette")
def create_app(
    *, pipeline: PipelineT, app_config: AppConfig, app_aiohttp_session: bool = True
) -> Tuple["fastapi.FastAPI", AppContext[PipelineT]]:
    @contextlib.asynccontextmanager
    async def _app_lifespan(app: "fastapi.FastAPI") -> AsyncGenerator[None, None]:
        ctx.pipeline = PipelineWrapper[PipelineT](pipeline)
        try:
            if app_aiohttp_session:
                async with aiohttp.ClientSession(
                    cookie_jar=aiohttp.DummyCookieJar()
                ) as aiohttp_session:
                    ctx.aiohttp_session = aiohttp_session
                    yield
            else:
                yield
        finally:
            await ctx.pipeline.close()

    # Should we control API versions?
    app = fastapi.FastAPI(lifespan=_app_lifespan)
    ctx = AppContext[PipelineT](config=app_config)
    app.state.context = ctx

    @app.get("/health", operation_id="checkHealth")
    async def _check_health() -> AIStudioNoResultResponse:
        return AIStudioNoResultResponse(
            logId=generate_log_id(), errorCode=0, errorMsg="Healthy"
        )

    @app.exception_handler(RequestValidationError)
    async def _validation_exception_handler(
        request: fastapi.Request, exc: RequestValidationError
    ) -> JSONResponse:
        json_compatible_data = jsonable_encoder(
            AIStudioNoResultResponse(
                logId=generate_log_id(),
                errorCode=422,
                errorMsg=json.dumps(exc.errors()),
            )
        )
        return JSONResponse(content=json_compatible_data, status_code=422)

    @app.exception_handler(HTTPException)
    async def _http_exception_handler(
        request: fastapi.Request, exc: HTTPException
    ) -> JSONResponse:
        json_compatible_data = jsonable_encoder(
            AIStudioNoResultResponse(
                logId=generate_log_id(), errorCode=exc.status_code, errorMsg=exc.detail
            )
        )
        return JSONResponse(content=json_compatible_data, status_code=exc.status_code)

    @app.exception_handler(Exception)
    async def _unexpected_exception_handler(
        request: fastapi.Request, exc: Exception
    ) -> JSONResponse:
        # XXX: The default server will duplicate the error message. Is it
        # necessary to log the exception info here?
        logging.exception("Unhandled exception")
        json_compatible_data = jsonable_encoder(
            AIStudioNoResultResponse(
                logId=generate_log_id(),
                errorCode=500,
                errorMsg="Internal server error",
            )
        )
        return JSONResponse(content=json_compatible_data, status_code=500)

    return app, ctx


# TODO: Precise type hints
@function_requires_deps("fastapi")
def primary_operation(
    app: "fastapi.FastAPI", path: str, operation_id: str, **kwargs: Any
) -> Callable:
    return app.post(
        path,
        operation_id=operation_id,
        responses={
            422: {"model": AIStudioNoResultResponse},
            500: {"model": AIStudioNoResultResponse},
        },
        response_model_exclude_none=True,
        **kwargs,
    )



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from ....utils.misc import abspath
from ..text_rec.config import TextRecConfig


class TableRecConfig(TextRecConfig):
    """Table Recognition Config"""

    def update_dataset(self, dataset_path: str, dataset_type: str = None):
        """update dataset settings

        Args:
            dataset_path (str): the root path of dataset.
            dataset_type (str, optional): dataset type. Defaults to None.

        Raises:
            ValueError: the dataset_type error.
        """
        dataset_path = abspath(dataset_path)
        if dataset_type is None:
            dataset_type = "PubTabTableRecDataset"
        if dataset_type == "PubTabTableRecDataset":
            _cfg = {
                "Train.dataset.name": dataset_type,
                "Train.dataset.data_dir": dataset_path,
                "Train.dataset.label_file_list": [
                    os.path.join(dataset_path, "train.txt")
                ],
                "Eval.dataset.name": dataset_type,
                "Eval.dataset.data_dir": dataset_path,
                "Eval.dataset.label_file_list": [os.path.join(dataset_path, "val.txt")],
            }
            self.update(_cfg)
        else:
            raise ValueError(f"{repr(dataset_type)} is not supported.")

    def _get_infer_shape(self):
        """get resize scale of ResizeImg operation in the evaluation

        Returns:
            str: resize scale, i.e. `Eval.dataset.transforms.ResizeTableImage.max_len`
        """
        size = None
        transforms = self.dict["Eval"]["dataset"]["transforms"]
        for op in transforms:
            if list(op)[0] == "ResizeTableImage":
                size = op["ResizeTableImage"]["max_len"]
        return size



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple, Union

import numpy as np

from ....modules.image_classification.model_list import MODELS
from ....utils.func_register import FuncRegister
from ...common.batch_sampler import ImageBatchSampler
from ...common.reader import ReadImage
from ..base import BasePredictor
from ..common import Normalize, Resize, ResizeByShort, ToBatch, ToCHWImage
from .processors import Crop, Topk
from .result import TopkResult


class ClasPredictor(BasePredictor):
    """ClasPredictor that inherits from BasePredictor."""

    entities = MODELS

    _FUNC_MAP = {}
    register = FuncRegister(_FUNC_MAP)

    def __init__(
        self, topk: Union[int, None] = None, *args: List, **kwargs: Dict
    ) -> None:
        """Initializes ClasPredictor.

        Args:
            topk (int, optional): The number of top-k predictions to return. If None, it will be depending on config of inference or predict. Defaults to None.
            *args: Arbitrary positional arguments passed to the superclass.
            **kwargs: Arbitrary keyword arguments passed to the superclass.
        """
        super().__init__(*args, **kwargs)
        self.topk = topk
        self.preprocessors, self.infer, self.postprocessors = self._build()

    def _build_batch_sampler(self) -> ImageBatchSampler:
        """Builds and returns an ImageBatchSampler instance.

        Returns:
            ImageBatchSampler: An instance of ImageBatchSampler.
        """
        return ImageBatchSampler()

    def _get_result_class(self) -> type:
        """Returns the result class, TopkResult.

        Returns:
            type: The TopkResult class.
        """
        return TopkResult

    def _build(self) -> Tuple:
        """Build the preprocessors, inference engine, and postprocessors based on the configuration.

        Returns:
            tuple: A tuple containing the preprocessors, inference engine, and postprocessors.
        """
        preprocessors = {"Read": ReadImage(format="RGB")}
        for cfg in self.config["PreProcess"]["transform_ops"]:
            tf_key = list(cfg.keys())[0]
            func = self._FUNC_MAP[tf_key]
            args = cfg.get(tf_key, {})
            name, op = func(self, **args) if args else func(self)
            preprocessors[name] = op
        preprocessors["ToBatch"] = ToBatch()

        infer = self.create_static_infer()

        postprocessors = {}
        for key in self.config["PostProcess"]:
            func = self._FUNC_MAP.get(key)
            args = self.config["PostProcess"].get(key, {})
            name, op = func(self, **args) if args else func(self)
            postprocessors[name] = op
        return preprocessors, infer, postprocessors

    def process(
        self, batch_data: List[Union[str, np.ndarray]], topk: Union[int, None] = None
    ) -> Dict[str, Any]:
        """
        Process a batch of data through the preprocessing, inference, and postprocessing.

        Args:
            batch_data (List[Union[str, np.ndarray], ...]): A batch of input data (e.g., image file paths).
            topk: The number of top predictions to keep. If None, it will be depending on `self.topk`. Defaults to None.

        Returns:
            dict: A dictionary containing the input path, raw image, class IDs, scores, and label names for every instance of the batch. Keys include 'input_path', 'input_img', 'class_ids', 'scores', and 'label_names'.
        """
        batch_raw_imgs = self.preprocessors["Read"](imgs=batch_data.instances)
        batch_imgs = self.preprocessors["Resize"](imgs=batch_raw_imgs)
        if "Crop" in self.preprocessors:
            batch_imgs = self.preprocessors["Crop"](imgs=batch_imgs)
        batch_imgs = self.preprocessors["Normalize"](imgs=batch_imgs)
        batch_imgs = self.preprocessors["ToCHW"](imgs=batch_imgs)
        x = self.preprocessors["ToBatch"](imgs=batch_imgs)
        batch_preds = self.infer(x=x)
        batch_class_ids, batch_scores, batch_label_names = self.postprocessors["Topk"](
            batch_preds, topk=topk or self.topk
        )
        return {
            "input_path": batch_data.input_paths,
            "page_index": batch_data.page_indexes,
            "input_img": batch_raw_imgs,
            "class_ids": batch_class_ids,
            "scores": batch_scores,
            "label_names": batch_label_names,
        }

    @register("ResizeImage")
    # TODO(gaotingquan): backend & interpolation
    def build_resize(
        self, resize_short=None, size=None, backend="cv2", interpolation="LINEAR"
    ):
        assert resize_short or size
        if resize_short:
            op = ResizeByShort(
                target_short_edge=resize_short,
                size_divisor=None,
                interp=interpolation,
                backend=backend,
            )
        else:
            op = Resize(
                target_size=size,
                size_divisor=None,
                interp=interpolation,
                backend=backend,
            )
        return "Resize", op

    @register("CropImage")
    def build_crop(self, size=224):
        return "Crop", Crop(crop_size=size)

    @register("NormalizeImage")
    def build_normalize(
        self,
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225],
        scale=1 / 255,
        order="",
        channel_num=3,
    ):
        assert channel_num == 3
        assert order == ""
        return "Normalize", Normalize(scale=scale, mean=mean, std=std)

    @register("ToCHWImage")
    def build_to_chw(self):
        return "ToCHW", ToCHWImage()

    @register("Topk")
    def build_topk(self, topk, label_list=None):
        if not self.topk:
            self.topk = int(topk)
        return "Topk", Topk(class_ids=label_list)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ....utils import logging
from .base_batch_sampler import BaseBatchSampler


class DocVLMBatchSampler(BaseBatchSampler):

    model_names_only_supports_batchsize_of_one = {"PP-DocBee-2B", "PP-DocBee-7B"}

    def __init__(self, model_name, batch_size: int = 1) -> None:
        """Initializes the BaseBatchSampler.

        Args:
            model_name (str): The name of the model.
            batch_size (int, optional): The size of each batch. Only support 1.
        """
        self.model_name = model_name
        if (
            self.model_name in self.model_names_only_supports_batchsize_of_one
            and batch_size != 1
        ):
            logging.warning(
                f"doc vlm batch sampler only support batch size 1 for {self.model_name}, but got {batch_size} and it will not take effect."
            )
            batch_size = 1
        super().__init__(batch_size)

    def sample(self, inputs):
        """Generate list of input file path.

        Args:
            inputs (str): file path.

        Yields:
            list: list of file path.
        """
        if isinstance(inputs, dict):
            inputs = [inputs]
        if not (isinstance(inputs, list) and all(isinstance(i, dict) for i in inputs)):
            raise TypeError(
                f"Not supported input data type! Only `Dict` or `List[Dict]` are supported, but got: {type(inputs)}."
            )

        batch = []
        for input_ in inputs:
            batch.append(input_)
            if len(batch) == self.batch_size:
                yield batch
                batch = []

        if len(batch) > 0:
            yield batch

    @BaseBatchSampler.batch_size.setter
    def batch_size(self, batch_size):
        """Sets the batch size.

        Args:
            batch_size (int): The batch size to set.

        Raises:
            Warning: If the batch size is not equal 1.
        """
        # only support batch size 1
        if (
            self.model_name in self.model_names_only_supports_batchsize_of_one
            and batch_size != 1
        ):
            logging.warning(
                f"doc vlm batch sampler only support batch size 1 for {self.model_name}, but got {batch_size} and it will not take effect."
            )
        else:
            self._batch_size = batch_size



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
import tarfile
from pathlib import Path
from typing import Any, Dict, List, Union

from ....utils import logging
from ....utils.cache import CACHE_DIR
from ....utils.download import download
from .base_batch_sampler import BaseBatchSampler


class Det3DBatchSampler(BaseBatchSampler):

    def __init__(self, temp_dir) -> None:
        super().__init__()
        self.temp_dir = temp_dir

    # XXX: auto download for url
    def _download_from_url(self, in_path: str) -> str:
        file_name = Path(in_path).name
        save_path = Path(CACHE_DIR) / "predict_input" / file_name
        download(in_path, save_path, overwrite=True)
        return save_path.as_posix()

    @property
    def batch_size(self) -> int:
        """Gets the batch size."""
        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size: int) -> None:
        """Sets the batch size.

        Args:
            batch_size (int): The batch size to set.
        """
        if batch_size != 1:
            logging.warning(
                "inference for 3D models only support batch_size equal to 1"
            )
        self._batch_size = batch_size

    def load_annotations(self, ann_file: str, data_root_dir: str) -> List[Dict]:
        """Load annotations from ann_file.

        Args:
            ann_file (str): Path of the annotation file.
            data_root_dir: (str): Path of the data root directory.

        Returns:
            list[dict]: List of annotations sorted by timestamps.
        """
        data = pickle.load(open(ann_file, "rb"))
        data_infos = list(sorted(data["infos"], key=lambda e: e["timestamp"]))
        # append root_dir to image and lidar filepaths
        for item in data_infos:
            # lidar data
            lidar_path = item["lidar_path"]
            new_lidar_path = os.path.join(data_root_dir, lidar_path)
            item["lidar_path"] = new_lidar_path
            # camera data
            cam_data = item["cams"]
            for cam_data_item_key in cam_data:
                cam_data_item = cam_data[cam_data_item_key]
                cam_data_item_path = cam_data_item["data_path"]
                new_cam_data_item_path = os.path.join(data_root_dir, cam_data_item_path)
                cam_data_item["data_path"] = new_cam_data_item_path
            # sweep data
            sweeps = item["sweeps"]
            for sweep_item in sweeps:
                sweep_item_path = sweep_item["data_path"]
                new_sweep_item_path = os.path.join(data_root_dir, sweep_item_path)
                sweep_item["data_path"] = new_sweep_item_path
        return data_infos

    def sample(self, inputs: Union[List[str], str]):
        if not isinstance(inputs, list):
            inputs = [inputs]

        sample_set = []
        for input in inputs:
            if isinstance(input, str):
                ann_path = (
                    self._download_from_url(input)
                    if input.startswith("http")
                    else input
                )
            else:
                logging.warning(
                    f"Not supported input data type! Only `str` is supported! So has been ignored: {input}."
                )
            # extract tar file to tempdir
            dataset_name = self.extract_tar(ann_path, self.temp_dir)
            data_root_dir = os.path.join(self.temp_dir, dataset_name)
            ann_pkl_path = os.path.join(data_root_dir, "nuscenes_infos_val.pkl")
            self.data_infos = self.load_annotations(ann_pkl_path, data_root_dir)
            sample_set.extend(self.data_infos)

        batch = []
        for sample in sample_set:
            batch.append(sample)
            if len(batch) == self.batch_size:
                yield batch
                batch = []

        if len(batch) > 0:
            yield batch

    def _rand_batch(self, data_size: int) -> List[Any]:
        raise NotImplementedError(
            "rand batch is not supported for 3D detection annotation data"
        )

    def extract_tar(self, tar_path, extract_path="."):
        try:
            memdirs = set()
            with tarfile.open(tar_path, "r") as tar:
                for member in tar.getmembers():
                    memdir = member.name.split("/")[0]
                    memdirs.add(memdir)
                    tar.extract(member, path=extract_path)
                logging.info(f"file extract to {extract_path}")
            assert (
                len(memdirs) == 1
            ), "Only one base directory is allowed for 3d bev dataset!"
            return list(memdirs)[0]
        except Exception as e:
            logging.error(f"error occurred while extracting tar file")
            raise e



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

import numpy as np

from ..layout_objects import LayoutBlock, LayoutRegion
from ..setting import BLOCK_LABEL_MAP, XYCUT_SETTINGS
from ..utils import (
    calculate_overlap_ratio,
    calculate_projection_overlap_ratio,
    get_seg_flag,
)


def get_nearest_edge_distance(
    bbox1: List[int],
    bbox2: List[int],
    weight: List[float] = [1.0, 1.0, 1.0, 1.0],
) -> Tuple[float]:
    """
    Calculate the nearest edge distance between two bounding boxes, considering directional weights.

    Args:
        bbox1 (list): The bounding box coordinates [x1, y1, x2, y2] of the input object.
        bbox2 (list): The bounding box coordinates [x1', y1', x2', y2'] of the object to match against.
        weight (list, optional): directional weights for the edge distances [left, right, up, down]. Defaults to [1, 1, 1, 1].

    Returns:
        float: The calculated minimum edge distance between the bounding boxes.
    """
    x1, y1, x2, y2 = bbox1
    x1_prime, y1_prime, x2_prime, y2_prime = bbox2
    min_x_distance, min_y_distance = 0, 0
    horizontal_iou = calculate_projection_overlap_ratio(bbox1, bbox2, "horizontal")
    vertical_iou = calculate_projection_overlap_ratio(bbox1, bbox2, "vertical")
    if horizontal_iou > 0 and vertical_iou > 0:
        return 0.0
    if horizontal_iou == 0:
        min_x_distance = min(abs(x1 - x2_prime), abs(x2 - x1_prime)) * (
            weight[0] if x2 < x1_prime else weight[1]
        )
    if vertical_iou == 0:
        min_y_distance = min(abs(y1 - y2_prime), abs(y2 - y1_prime)) * (
            weight[2] if y2 < y1_prime else weight[3]
        )

    return min_x_distance + min_y_distance


def projection_by_bboxes(boxes: np.ndarray, axis: int) -> np.ndarray:
    """
    Generate a 1D projection histogram from bounding boxes along a specified axis.

    Args:
        boxes: A (N, 4) array of bounding boxes defined by [x_min, y_min, x_max, y_max].
        axis: Axis for projection; 0 for horizontal (x-axis), 1 for vertical (y-axis).

    Returns:
        A 1D numpy array representing the projection histogram based on bounding box intervals.
    """
    assert axis in [0, 1]

    if np.min(boxes[:, axis::2]) < 0:
        max_length = abs(np.min(boxes[:, axis::2]))
    else:
        max_length = np.max(boxes[:, axis::2])

    projection = np.zeros(max_length, dtype=int)

    # Increment projection histogram over the interval defined by each bounding box
    for start, end in boxes[:, axis::2]:
        start = abs(start)
        end = abs(end)
        projection[start:end] += 1

    return projection


def split_projection_profile(arr_values: np.ndarray, min_value: float, min_gap: float):
    """
    Split the projection profile into segments based on specified thresholds.

    Args:
        arr_values: 1D array representing the projection profile.
        min_value: Minimum value threshold to consider a profile segment significant.
        min_gap: Minimum gap width to consider a separation between segments.

    Returns:
        A tuple of start and end indices for each segment that meets the criteria.
    """
    # Identify indices where the projection exceeds the minimum value
    significant_indices = np.where(arr_values > min_value)[0]
    if not len(significant_indices):
        return

    # Calculate gaps between significant indices
    index_diffs = significant_indices[1:] - significant_indices[:-1]
    gap_indices = np.where(index_diffs > min_gap)[0]

    # Determine start and end indices of segments
    segment_starts = np.insert(
        significant_indices[gap_indices + 1],
        0,
        significant_indices[0],
    )
    segment_ends = np.append(
        significant_indices[gap_indices],
        significant_indices[-1] + 1,
    )

    return segment_starts, segment_ends


def recursive_yx_cut(
    boxes: np.ndarray, indices: List[int], res: List[int], min_gap: int = 1
):
    """
    Recursively project and segment bounding boxes, starting with Y-axis and followed by X-axis.

    Args:
        boxes: A (N, 4) array representing bounding boxes.
        indices: List of indices indicating the original position of boxes.
        res: List to store indices of the final segmented bounding boxes.
        min_gap (int): Minimum gap width to consider a separation between segments on the X-axis. Defaults to 1.

    Returns:
        None: This function modifies the `res` list in place.
    """
    assert len(boxes) == len(
        indices
    ), "The length of boxes and indices must be the same."

    # Sort by y_min for Y-axis projection
    y_sorted_indices = boxes[:, 1].argsort()
    y_sorted_boxes = boxes[y_sorted_indices]
    y_sorted_indices = np.array(indices)[y_sorted_indices]

    # Perform Y-axis projection
    y_projection = projection_by_bboxes(boxes=y_sorted_boxes, axis=1)
    y_intervals = split_projection_profile(y_projection, 0, 1)

    if not y_intervals:
        return

    # Process each segment defined by Y-axis projection
    for y_start, y_end in zip(*y_intervals):
        # Select boxes within the current y interval
        y_interval_indices = (y_start <= y_sorted_boxes[:, 1]) & (
            y_sorted_boxes[:, 1] < y_end
        )
        y_boxes_chunk = y_sorted_boxes[y_interval_indices]
        y_indices_chunk = y_sorted_indices[y_interval_indices]

        # Sort by x_min for X-axis projection
        x_sorted_indices = y_boxes_chunk[:, 0].argsort()
        x_sorted_boxes_chunk = y_boxes_chunk[x_sorted_indices]
        x_sorted_indices_chunk = y_indices_chunk[x_sorted_indices]

        # Perform X-axis projection
        x_projection = projection_by_bboxes(boxes=x_sorted_boxes_chunk, axis=0)
        x_intervals = split_projection_profile(x_projection, 0, min_gap)

        if not x_intervals:
            continue

        # If X-axis cannot be further segmented, add current indices to results
        if len(x_intervals[0]) == 1:
            res.extend(x_sorted_indices_chunk)
            continue

        if np.min(x_sorted_boxes_chunk[:, 0]) < 0:
            x_intervals = np.flip(x_intervals, axis=1)
        # Recursively process each segment defined by X-axis projection
        for x_start, x_end in zip(*x_intervals):
            x_interval_indices = (x_start <= abs(x_sorted_boxes_chunk[:, 0])) & (
                abs(x_sorted_boxes_chunk[:, 0]) < x_end
            )
            recursive_yx_cut(
                x_sorted_boxes_chunk[x_interval_indices],
                x_sorted_indices_chunk[x_interval_indices],
                res,
            )


def recursive_xy_cut(
    boxes: np.ndarray, indices: List[int], res: List[int], min_gap: int = 1
):
    """
    Recursively performs X-axis projection followed by Y-axis projection to segment bounding boxes.

    Args:
        boxes: A (N, 4) array representing bounding boxes with [x_min, y_min, x_max, y_max].
        indices: A list of indices representing the position of boxes in the original data.
        res: A list to store indices of bounding boxes that meet the criteria.
        min_gap (int): Minimum gap width to consider a separation between segments on the X-axis. Defaults to 1.

    Returns:
        None: This function modifies the `res` list in place.
    """
    # Ensure boxes and indices have the same length
    assert len(boxes) == len(
        indices
    ), "The length of boxes and indices must be the same."

    # Sort by x_min to prepare for X-axis projection
    x_sorted_indices = boxes[:, 0].argsort()
    x_sorted_boxes = boxes[x_sorted_indices]
    x_sorted_indices = np.array(indices)[x_sorted_indices]

    # Perform X-axis projection
    x_projection = projection_by_bboxes(boxes=x_sorted_boxes, axis=0)
    x_intervals = split_projection_profile(x_projection, 0, 1)

    if not x_intervals:
        return

    if np.min(x_sorted_boxes[:, 0]) < 0:
        x_intervals = np.flip(x_intervals, axis=1)
    # Process each segment defined by X-axis projection
    for x_start, x_end in zip(*x_intervals):
        # Select boxes within the current x interval
        x_interval_indices = (x_start <= abs(x_sorted_boxes[:, 0])) & (
            abs(x_sorted_boxes[:, 0]) < x_end
        )
        x_boxes_chunk = x_sorted_boxes[x_interval_indices]
        x_indices_chunk = x_sorted_indices[x_interval_indices]

        # Sort selected boxes by y_min to prepare for Y-axis projection
        y_sorted_indices = x_boxes_chunk[:, 1].argsort()
        y_sorted_boxes_chunk = x_boxes_chunk[y_sorted_indices]
        y_sorted_indices_chunk = x_indices_chunk[y_sorted_indices]

        # Perform Y-axis projection
        y_projection = projection_by_bboxes(boxes=y_sorted_boxes_chunk, axis=1)
        y_intervals = split_projection_profile(y_projection, 0, min_gap)

        if not y_intervals:
            continue

        # If Y-axis cannot be further segmented, add current indices to results
        if len(y_intervals[0]) == 1:
            res.extend(y_sorted_indices_chunk)
            continue

        # Recursively process each segment defined by Y-axis projection
        for y_start, y_end in zip(*y_intervals):
            y_interval_indices = (y_start <= y_sorted_boxes_chunk[:, 1]) & (
                y_sorted_boxes_chunk[:, 1] < y_end
            )
            recursive_xy_cut(
                y_sorted_boxes_chunk[y_interval_indices],
                y_sorted_indices_chunk[y_interval_indices],
                res,
            )


def reference_insert(
    block: LayoutBlock,
    sorted_blocks: List[LayoutBlock],
    **kwargs,
):
    """
    Insert reference block into sorted blocks based on the distance between the block and the nearest sorted block.

    Args:
        block: The block to insert into the sorted blocks.
        sorted_blocks: The sorted blocks where the new block will be inserted.
        config: Configuration dictionary containing parameters related to the layout parsing.
        median_width: Median width of the document. Defaults to 0.0.

    Returns:
        sorted_blocks: The updated sorted blocks after insertion.
    """
    min_distance = float("inf")
    nearest_sorted_block_index = 0
    for sorted_block_idx, sorted_block in enumerate(sorted_blocks):
        if sorted_block.bbox[3] <= block.bbox[1]:
            distance = -(sorted_block.bbox[2] * 10 + sorted_block.bbox[3])
        if distance < min_distance:
            min_distance = distance
            nearest_sorted_block_index = sorted_block_idx

    sorted_blocks.insert(nearest_sorted_block_index + 1, block)
    return sorted_blocks


def manhattan_insert(
    block: LayoutBlock,
    sorted_blocks: List[LayoutBlock],
    **kwargs,
):
    """
    Insert a block into a sorted list of blocks based on the Manhattan distance between the block and the nearest sorted block.

    Args:
        block: The block to insert into the sorted blocks.
        sorted_blocks: The sorted blocks where the new block will be inserted.
        config: Configuration dictionary containing parameters related to the layout parsing.
        median_width: Median width of the document. Defaults to 0.0.

    Returns:
        sorted_blocks: The updated sorted blocks after insertion.
    """
    min_distance = float("inf")
    nearest_sorted_block_index = 0
    for sorted_block_idx, sorted_block in enumerate(sorted_blocks):
        distance = _manhattan_distance(block.bbox, sorted_block.bbox)
        if distance < min_distance:
            min_distance = distance
            nearest_sorted_block_index = sorted_block_idx

    sorted_blocks.insert(nearest_sorted_block_index + 1, block)
    return sorted_blocks


def euclidean_insert(
    block: LayoutRegion,
    sorted_blocks: List[LayoutRegion],
    **kwargs,
):
    """
    Insert a block into a sorted list of blocks based on the Euclidean distance between the block and the nearest sorted block.

    Args:
        block: The block to insert into the sorted blocks.
        sorted_blocks: The sorted blocks where the new block will be inserted.
        config: Configuration dictionary containing parameters related to the layout parsing.
        median_width: Median width of the document. Defaults to 0.0.

    Returns:
        sorted_blocks: The updated sorted blocks after insertion.
    """
    nearest_sorted_block_index = len(sorted_blocks)
    block_euclidean_distance = block.euclidean_distance
    for sorted_block_idx, sorted_block in enumerate(sorted_blocks):
        distance = sorted_block.euclidean_distance
        if distance > block_euclidean_distance:
            nearest_sorted_block_index = sorted_block_idx
            break
    sorted_blocks.insert(nearest_sorted_block_index, block)
    return sorted_blocks


def weighted_distance_insert(
    block: LayoutBlock,
    sorted_blocks: List[LayoutBlock],
    region: LayoutRegion,
):
    """
    Insert a block into a sorted list of blocks based on the weighted distance between the block and the nearest sorted block.

    Args:
        block: The block to insert into the sorted blocks.
        sorted_blocks: The sorted blocks where the new block will be inserted.
        config: Configuration dictionary containing parameters related to the layout parsing.
        median_width: Median width of the document. Defaults to 0.0.

    Returns:
        sorted_blocks: The updated sorted blocks after insertion.
    """

    tolerance_len = XYCUT_SETTINGS["edge_distance_compare_tolerance_len"]
    x1, y1, x2, y2 = block.bbox
    min_weighted_distance, min_edge_distance, min_up_edge_distance = (
        float("inf"),
        float("inf"),
        float("inf"),
    )
    nearest_sorted_block_index = 0
    for sorted_block_idx, sorted_block in enumerate(sorted_blocks):

        x1_prime, y1_prime, x2_prime, y2_prime = sorted_block.bbox

        # Calculate edge distance
        weight = _get_weights(block.order_label, block.direction)
        edge_distance = get_nearest_edge_distance(block.bbox, sorted_block.bbox, weight)

        if block.label in BLOCK_LABEL_MAP["doc_title_labels"]:
            disperse = max(1, region.text_line_width)
            tolerance_len = max(tolerance_len, disperse)
        if block.label == "abstract":
            tolerance_len *= 2
            edge_distance = max(0.1, edge_distance) * 10

        # Calculate up edge distances
        up_edge_distance = y1_prime if region.direction == "horizontal" else -x2_prime
        left_edge_distance = x1_prime if region.direction == "horizontal" else y1_prime
        is_below_sorted_block = (
            y2_prime < y1 if region.direction == "horizontal" else x1_prime > x2
        )

        if (
            block.label not in BLOCK_LABEL_MAP["unordered_labels"]
            or block.label in BLOCK_LABEL_MAP["doc_title_labels"]
            or block.label in BLOCK_LABEL_MAP["paragraph_title_labels"]
            or block.label in BLOCK_LABEL_MAP["vision_labels"]
        ) and is_below_sorted_block:
            up_edge_distance = -up_edge_distance
            left_edge_distance = -left_edge_distance

        if abs(min_up_edge_distance - up_edge_distance) <= tolerance_len:
            up_edge_distance = min_up_edge_distance

        # Calculate weighted distance
        weighted_distance = (
            +edge_distance
            * XYCUT_SETTINGS["distance_weight_map"].get("edge_weight", 10**4)
            + up_edge_distance
            * XYCUT_SETTINGS["distance_weight_map"].get("up_edge_weight", 1)
            + left_edge_distance
            * XYCUT_SETTINGS["distance_weight_map"].get("left_edge_weight", 0.0001)
        )

        min_edge_distance = min(edge_distance, min_edge_distance)
        min_up_edge_distance = min(up_edge_distance, min_up_edge_distance)

        if weighted_distance < min_weighted_distance:
            nearest_sorted_block_index = sorted_block_idx
            min_weighted_distance = weighted_distance
            if abs(y1 // 2 - y1_prime // 2) > 0:
                sorted_distance = y1_prime
                block_distance = y1
            else:
                if region.direction == "horizontal":
                    if abs(x1 // 2 - x2 // 2) > 0:
                        sorted_distance = x1_prime
                        block_distance = x1
                    else:
                        # distance with (0,0)
                        sorted_block_center_x, sorted_block_center_y = (
                            sorted_block.get_centroid()
                        )
                        block_center_x, block_center_y = block.get_centroid()
                        sorted_distance = (
                            sorted_block_center_x**2 + sorted_block_center_y**2
                        )
                        block_distance = block_center_x**2 + block_center_y**2
                else:
                    if abs(x1 - x2) > 0:
                        sorted_distance = -x2_prime
                        block_distance = -x2
                    else:
                        # distance with (max,0)
                        sorted_block_center_x, sorted_block_center_y = (
                            sorted_block.get_centroid()
                        )
                        block_center_x, block_center_y = block.get_centroid()
                        sorted_distance = (
                            sorted_block_center_x**2 + sorted_block_center_y**2
                        )
                        block_distance = block_center_x**2 + block_center_y**2
            if block_distance > sorted_distance:
                nearest_sorted_block_index = sorted_block_idx + 1
                if (
                    sorted_block_idx < len(sorted_blocks) - 1
                    and block.label
                    in BLOCK_LABEL_MAP["vision_labels"]
                    + BLOCK_LABEL_MAP["vision_title_labels"]
                ):
                    seg_start_flag, _ = get_seg_flag(
                        sorted_blocks[sorted_block_idx + 1],
                        sorted_blocks[sorted_block_idx],
                    )
                    if not seg_start_flag:
                        nearest_sorted_block_index += 1
            else:
                if (
                    sorted_block_idx > 0
                    and block.label
                    in BLOCK_LABEL_MAP["vision_labels"]
                    + BLOCK_LABEL_MAP["vision_title_labels"]
                ):
                    seg_start_flag, _ = get_seg_flag(
                        sorted_blocks[sorted_block_idx],
                        sorted_blocks[sorted_block_idx - 1],
                    )
                    if not seg_start_flag:
                        nearest_sorted_block_index = sorted_block_idx - 1

    sorted_blocks.insert(nearest_sorted_block_index, block)
    return sorted_blocks


def insert_child_blocks(
    block: LayoutBlock,
    block_idx: int,
    sorted_blocks: List[LayoutBlock],
) -> List[LayoutBlock]:
    """
    Insert child blocks of a block into the sorted blocks list.

    Args:
        block: The parent block whose child blocks need to be inserted.
        block_idx: Index at which the parent block exists in the sorted blocks list.
        sorted_blocks: Sorted blocks list where the child blocks are to be inserted.

    Returns:
        sorted_blocks: Updated sorted blocks list after inserting child blocks.
    """
    if block.child_blocks:
        sub_blocks = block.get_child_blocks()
        sub_blocks.append(block)
        sub_blocks = sort_child_blocks(sub_blocks, sub_blocks[0].direction)
        sorted_blocks[block_idx] = sub_blocks[0]
        for block in sub_blocks[1:]:
            block_idx += 1
            sorted_blocks.insert(block_idx, block)
    return sorted_blocks


def sort_child_blocks(
    blocks: List[LayoutRegion], direction="horizontal"
) -> List[LayoutBlock]:
    """
    Sort child blocks based on their bounding box coordinates.

    Args:
        blocks: A list of LayoutBlock objects representing the child blocks.
        direction: direction of the blocks ('horizontal' or 'vertical'). Default is 'horizontal'.
    Returns:
        sorted_blocks: A sorted list of LayoutBlock objects.
    """
    if blocks[0].label != "region":
        if direction == "horizontal":
            blocks.sort(
                key=lambda x: (
                    x.bbox[1],
                    x.bbox[0],
                    x.get_centroid()[0] ** 2 + x.get_centroid()[1] ** 2,
                ),  # distance with (0,0)
            )
        else:
            blocks.sort(
                key=lambda x: (
                    -x.bbox[2],
                    x.bbox[1],
                    -x.get_centroid()[0] ** 2 + x.get_centroid()[1] ** 2,
                ),  # distance with (max,0)
            )
    else:
        blocks.sort(key=lambda x: x.euclidean_distance)
    return blocks


def _get_weights(label, direction="horizontal"):
    """Define weights based on the label and direction."""
    if label == "doc_title":
        return (
            [1, 0.1, 0.1, 1] if direction == "horizontal" else [0.2, 0.1, 1, 1]
        )  # left-down ,  right-left
    elif label in [
        "paragraph_title",
        "table_title",
        "abstract",
        "image",
        "seal",
        "chart",
        "figure",
    ]:
        return [1, 1, 0.1, 1]  # down
    else:
        return [1, 1, 1, 0.1]  # up


def _manhattan_distance(
    point1: Tuple[float, float],
    point2: Tuple[float, float],
    weight_x: float = 1.0,
    weight_y: float = 1.0,
) -> float:
    """
    Calculate the weighted Manhattan distance between two points.

    Args:
        point1 (Tuple[float, float]): The first point as (x, y).
        point2 (Tuple[float, float]): The second point as (x, y).
        weight_x (float): The weight for the x-axis distance. Default is 1.0.
        weight_y (float): The weight for the y-axis distance. Default is 1.0.

    Returns:
        float: The weighted Manhattan distance between the two points.
    """
    return weight_x * abs(point1[0] - point2[0]) + weight_y * abs(point1[1] - point2[1])


def sort_normal_blocks(
    blocks, text_line_height, text_line_width, region_direction
) -> List[LayoutBlock]:
    """Sort blocks by their position within the page

    Args:
        blocks (List[LayoutBlock]): List of blocks to be sorted.
        text_line_height (int): Height of each line of text.
        text_line_width (int): Width of each line of text.
        region_direction (str): Direction of the region, either "horizontal" or "vertical".

    Returns:
        List[LayoutBlock]: Sorted list of blocks.
    """
    if region_direction == "horizontal":
        blocks.sort(
            key=lambda x: (
                x.bbox[1] // text_line_height,
                x.bbox[0] // text_line_width,
                x.get_centroid()[0] ** 2 + x.get_centroid()[1] ** 2,
            ),
        )
    else:
        blocks.sort(
            key=lambda x: (
                -x.bbox[2] // text_line_width,
                x.bbox[1] // text_line_height,
                -x.get_centroid()[0] ** 2 + x.get_centroid()[1] ** 2,
            ),
        )
    return blocks


def get_cut_blocks(blocks, cut_direction, cut_coordinates, mask_labels=[]):
    """
    Cut blocks based on the given cut direction and coordinates.

    Args:
        blocks (list): list of blocks to be cut.
        cut_direction (str): cut direction, either "horizontal" or "vertical".
        cut_coordinates (list): list of cut coordinates.

    Returns:
        list: a list of tuples containing the cutted blocks and their corresponding mean width。
    """
    cuted_list = []
    # filter out mask blocks,including header, footer, unordered and child_blocks

    # 0: horizontal, 1: vertical
    cut_aixis = 0 if cut_direction == "horizontal" else 1
    blocks.sort(key=lambda x: x.bbox[cut_aixis + 2])
    cut_coordinates.append(float("inf"))

    cut_coordinates = list(set(cut_coordinates))
    cut_coordinates.sort()

    cut_idx = 0
    for cut_coordinate in cut_coordinates:
        group_blocks = []
        block_idx = cut_idx
        while block_idx < len(blocks):
            block = blocks[block_idx]
            if block.bbox[cut_aixis + 2] > cut_coordinate:
                break
            elif block.order_label not in mask_labels:
                group_blocks.append(block)
            block_idx += 1
        cut_idx = block_idx
        if group_blocks:
            cuted_list.append(group_blocks)

    return cuted_list


def get_blocks_by_direction_interval(
    blocks: List[LayoutBlock],
    start_index: int,
    end_index: int,
    direction: str = "horizontal",
) -> List[LayoutBlock]:
    """
    Get blocks within a specified direction interval.

    Args:
        blocks (List[LayoutBlock]): A list of blocks.
        start_index (int): The starting index of the direction.
        end_index (int): The ending index of the direction.
        direction (str, optional): The direction to consider. Defaults to "horizontal".

    Returns:
        List[LayoutBlock]: A list of blocks within the specified direction interval.
    """
    interval_blocks = []
    aixis = 0 if direction == "horizontal" else 1
    blocks.sort(key=lambda x: x.bbox[aixis + 2])

    for block in blocks:
        if block.bbox[aixis] >= start_index and block.bbox[aixis + 2] <= end_index:
            interval_blocks.append(block)

    return interval_blocks


def get_nearest_blocks(
    block: LayoutBlock,
    ref_blocks: List[LayoutBlock],
    overlap_threshold,
    direction="horizontal",
) -> List:
    """
    Get the adjacent blocks with the same direction as the current block.
    Args:
        block (LayoutBlock): The current block.
        blocks (List[LayoutBlock]): A list of all blocks.
        ref_block_idxes (List[int]): A list of indices of reference blocks.
        iou_threshold (float): The IOU threshold to determine if two blocks are considered adjacent.
    Returns:
        Int: The index of the previous block with same direction.
        Int: The index of the following block with same direction.
    """
    prev_blocks: List[LayoutBlock] = []
    post_blocks: List[LayoutBlock] = []
    sort_index = 1 if direction == "horizontal" else 0
    for ref_block in ref_blocks:
        if ref_block.index == block.index:
            continue
        overlap_ratio = calculate_projection_overlap_ratio(
            block.bbox, ref_block.bbox, direction, mode="small"
        )
        if overlap_ratio > overlap_threshold:
            if ref_block.bbox[sort_index] <= block.bbox[sort_index]:
                prev_blocks.append(ref_block)
            else:
                post_blocks.append(ref_block)

    if prev_blocks:
        prev_blocks.sort(key=lambda x: x.bbox[sort_index], reverse=True)
    if post_blocks:
        post_blocks.sort(key=lambda x: x.bbox[sort_index])

    return prev_blocks, post_blocks


def update_doc_title_child_blocks(
    block: LayoutBlock,
    region: LayoutRegion,
) -> None:
    """
    Update the child blocks of a document title block.

    The child blocks need to meet the following conditions:
        1. They must be adjacent
        2. They must have the same direction as the parent block.
        3. Their short side length should be less than 80% of the parent's short side length.
        4. Their long side length should be less than 150% of the parent's long side length.
        5. The child block must be text block.
        6. The nearest edge distance should be less than 2 times of the text line height.

    Args:
        blocks (List[LayoutBlock]): overall blocks.
        block (LayoutBlock): document title block.
        prev_idx (int): previous block index, None if not exist.
        post_idx (int): post block index, None if not exist.
        config (dict): configurations.

    Returns:
        None

    """
    ref_blocks = [region.block_map[idx] for idx in region.normal_text_block_idxes]
    overlap_threshold = XYCUT_SETTINGS["child_block_overlap_ratio_threshold"]
    prev_blocks, post_blocks = get_nearest_blocks(
        block, ref_blocks, overlap_threshold, block.direction
    )
    prev_block = None
    post_block = None

    if prev_blocks:
        prev_block = prev_blocks[0]
    if post_blocks:
        post_block = post_blocks[0]

    for ref_block in [prev_block, post_block]:
        if ref_block is None:
            continue
        with_seem_direction = ref_block.direction == block.direction

        short_side_length_condition = (
            ref_block.short_side_length < block.short_side_length * 0.8
        )

        long_side_length_condition = (
            ref_block.long_side_length < block.long_side_length
            or ref_block.long_side_length > 1.5 * block.long_side_length
        )

        nearest_edge_distance = get_nearest_edge_distance(block.bbox, ref_block.bbox)

        if (
            with_seem_direction
            and ref_block.label in BLOCK_LABEL_MAP["text_labels"]
            and short_side_length_condition
            and long_side_length_condition
            and ref_block.num_of_lines < 3
            and nearest_edge_distance < ref_block.text_line_height * 2
        ):
            ref_block.order_label = "doc_title_text"
            block.append_child_block(ref_block)
            region.normal_text_block_idxes.remove(ref_block.index)

    for ref_block in ref_blocks:
        if ref_block.order_label == "doc_title_text":
            continue
        with_seem_direction = ref_block.direction == block.direction

        overlap_ratio = calculate_overlap_ratio(
            block.bbox, ref_block.bbox, mode="small"
        )

        if overlap_ratio > 0.9 and with_seem_direction:
            ref_block.order_label = "doc_title_text"
            block.append_child_block(ref_block)
            region.normal_text_block_idxes.remove(ref_block.index)


def update_paragraph_title_child_blocks(
    block: LayoutBlock,
    region: LayoutRegion,
) -> None:
    """
    Update the child blocks of a paragraph title block.

    The child blocks need to meet the following conditions:
        1. They must be adjacent
        2. They must have the same direction as the parent block.
        3. The child block must be paragraph title block.

    Args:
        blocks (List[LayoutBlock]): overall blocks.
        block (LayoutBlock): document title block.
        prev_idx (int): previous block index, None if not exist.
        post_idx (int): post block index, None if not exist.
        config (dict): configurations.

    Returns:
        None

    """
    if block.order_label == "sub_paragraph_title":
        return
    ref_blocks = [
        region.block_map[idx]
        for idx in region.paragraph_title_block_idxes + region.normal_text_block_idxes
    ]
    overlap_threshold = XYCUT_SETTINGS["child_block_overlap_ratio_threshold"]
    prev_blocks, post_blocks = get_nearest_blocks(
        block, ref_blocks, overlap_threshold, block.direction
    )
    for ref_blocks in [prev_blocks, post_blocks]:
        for ref_block in ref_blocks:
            if ref_block.label not in BLOCK_LABEL_MAP["paragraph_title_labels"]:
                break
            min_text_line_height = min(
                block.text_line_height, ref_block.text_line_height
            )
            nearest_edge_distance = get_nearest_edge_distance(
                block.bbox, ref_block.bbox
            )
            with_seem_direction = ref_block.direction == block.direction
            with_seem_start = (
                abs(ref_block.start_coordinate - block.start_coordinate)
                < min_text_line_height * 2
            )
            if (
                with_seem_direction
                and with_seem_start
                and nearest_edge_distance <= min_text_line_height * 1.5
            ):
                ref_block.order_label = "sub_paragraph_title"
                block.append_child_block(ref_block)
                region.paragraph_title_block_idxes.remove(ref_block.index)


def update_vision_child_blocks(
    block: LayoutBlock,
    region: LayoutRegion,
) -> None:
    """
    Update the child blocks of a paragraph title block.

    The child blocks need to meet the following conditions:
    - For Both:
        1. They must be adjacent
        2. The child block must be vision_title or text block.
    - For vision_title:
        1. The distance between the child block and the parent block should be less than 1/2 of the parent's height.
    - For text block:
        1. The distance between the child block and the parent block should be less than 15.
        2. The child short_side_length should be less than the parent's short side length.
        3. The child long_side_length should be less than 50% of the parent's long side length.
        4. The difference between their centers is very small.

    Args:
        blocks (List[LayoutBlock]): overall blocks.
        block (LayoutBlock): document title block.
        ref_block_idxes (List[int]): A list of indices of reference blocks.
        prev_idx (int): previous block index, None if not exist.
        post_idx (int): post block index, None if not exist.
        config (dict): configurations.

    Returns:
        None

    """
    ref_blocks = [
        region.block_map[idx]
        for idx in region.normal_text_block_idxes + region.vision_title_block_idxes
    ]
    overlap_threshold = XYCUT_SETTINGS["child_block_overlap_ratio_threshold"]
    has_vision_footnote = False
    has_vision_title = False
    for direction in [block.direction, block.secondary_direction]:
        prev_blocks, post_blocks = get_nearest_blocks(
            block, ref_blocks, overlap_threshold, direction
        )
        for ref_block in prev_blocks:
            if (
                ref_block.label
                not in BLOCK_LABEL_MAP["text_labels"]
                + BLOCK_LABEL_MAP["vision_title_labels"]
            ):
                break
            nearest_edge_distance = get_nearest_edge_distance(
                block.bbox, ref_block.bbox
            )
            block_center = block.get_centroid()
            ref_block_center = ref_block.get_centroid()
            if (
                ref_block.label in BLOCK_LABEL_MAP["vision_title_labels"]
                and nearest_edge_distance <= ref_block.text_line_height * 2
            ):
                has_vision_title = True
                ref_block.order_label = "vision_title"
                block.append_child_block(ref_block)
                region.vision_title_block_idxes.remove(ref_block.index)
            if ref_block.label in BLOCK_LABEL_MAP["text_labels"]:
                if (
                    not has_vision_footnote
                    and ref_block.direction == block.direction
                    and ref_block.long_side_length < block.long_side_length
                    and nearest_edge_distance <= ref_block.text_line_height * 2
                ):
                    if (
                        (
                            ref_block.short_side_length < block.short_side_length
                            and ref_block.long_side_length
                            < 0.5 * block.long_side_length
                            and abs(block_center[0] - ref_block_center[0]) < 10
                        )
                        or (
                            block.bbox[0] - ref_block.bbox[0] < 10
                            and ref_block.num_of_lines == 1
                        )
                        or (
                            block.bbox[2] - ref_block.bbox[2] < 10
                            and ref_block.num_of_lines == 1
                        )
                    ):
                        has_vision_footnote = True
                        ref_block.order_label = "vision_footnote"
                        block.append_child_block(ref_block)
                        region.normal_text_block_idxes.remove(ref_block.index)
                break
        for ref_block in post_blocks:
            if (
                has_vision_footnote
                and ref_block.label in BLOCK_LABEL_MAP["text_labels"]
            ):
                break
            nearest_edge_distance = get_nearest_edge_distance(
                block.bbox, ref_block.bbox
            )
            block_center = block.get_centroid()
            ref_block_center = ref_block.get_centroid()
            if (
                ref_block.label in BLOCK_LABEL_MAP["vision_title_labels"]
                and nearest_edge_distance <= ref_block.text_line_height * 2
            ):
                has_vision_title = True
                ref_block.order_label = "vision_title"
                block.append_child_block(ref_block)
                region.vision_title_block_idxes.remove(ref_block.index)
            if ref_block.label in BLOCK_LABEL_MAP["text_labels"]:
                if (
                    not has_vision_footnote
                    and ref_block.direction == block.direction
                    and ref_block.long_side_length < block.long_side_length
                    and nearest_edge_distance <= ref_block.text_line_height * 2
                ):
                    if (
                        (
                            ref_block.short_side_length < block.short_side_length
                            and ref_block.long_side_length
                            < 0.5 * block.long_side_length
                            and abs(block_center[0] - ref_block_center[0]) < 10
                        )
                        or (
                            block.bbox[0] - ref_block.bbox[0] < 10
                            and ref_block.num_of_lines == 1
                        )
                        or (
                            block.bbox[2] - ref_block.bbox[2] < 10
                            and ref_block.num_of_lines == 1
                        )
                    ):
                        has_vision_footnote = True
                        ref_block.label = "vision_footnote"
                        ref_block.order_label = "vision_footnote"
                        block.append_child_block(ref_block)
                        region.normal_text_block_idxes.remove(ref_block.index)
                break
        if has_vision_title:
            break

    for ref_block in ref_blocks:
        if ref_block.index not in region.normal_text_block_idxes:
            continue

        overlap_ratio = calculate_overlap_ratio(
            block.bbox, ref_block.bbox, mode="small"
        )

        if overlap_ratio > 0.9:
            ref_block.label = "vision_footnote"
            ref_block.order_label = "vision_footnote"
            block.append_child_block(ref_block)
            region.normal_text_block_idxes.remove(ref_block.index)


def update_region_child_blocks(
    block: LayoutBlock,
    region: LayoutRegion,
) -> None:
    """Update child blocks of a region.

    Args:
        block (LayoutBlock): document title block.
        region (LayoutRegion): layout region.

    Returns:
        None
    """
    for ref_block in region.block_map.values():
        if block.index != ref_block.index:
            bbox_iou = calculate_overlap_ratio(block.bbox, ref_block.bbox)
            if (
                bbox_iou > 0
                and block.area > ref_block.area
                and ref_block.order_label != "sub_region"
            ):
                ref_block.order_label = "sub_region"
                block.append_child_block(ref_block)
                region.normal_text_block_idxes.remove(ref_block.index)


def calculate_discontinuous_projection(
    boxes, direction="horizontal", return_num=False
) -> List:
    """
    Calculate the discontinuous projection of boxes along the specified direction.

    Args:
        boxes (ndarray): Array of bounding boxes represented by [[x_min, y_min, x_max, y_max]].
        direction (str): direction along which to perform the projection ('horizontal' or 'vertical').

    Returns:
        list: List of tuples representing the merged intervals.
    """
    boxes = np.array(boxes)
    if direction == "horizontal":
        intervals = boxes[:, [0, 2]]
    elif direction == "vertical":
        intervals = boxes[:, [1, 3]]
    else:
        raise ValueError("direction must be 'horizontal' or 'vertical'")

    intervals = intervals[np.argsort(intervals[:, 0])]

    merged_intervals = []
    num = 1
    current_start, current_end = intervals[0]
    num_list = []

    for start, end in intervals[1:]:
        if start <= current_end:
            num += 1
            current_end = max(current_end, end)
        else:
            num_list.append(num)
            merged_intervals.append((current_start, current_end))
            num = 1
            current_start, current_end = start, end

    num_list.append(num)
    merged_intervals.append((current_start, current_end))
    if return_num:
        return merged_intervals, num_list
    return merged_intervals


def shrink_overlapping_boxes(
    boxes, direction="horizontal", min_threshold=0, max_threshold=0.1
) -> List:
    """
    Shrink overlapping boxes along the specified direction.

    Args:
        boxes (ndarray): Array of bounding boxes represented by [[x_min, y_min, x_max, y_max]].
        direction (str): direction along which to perform the shrinking ('horizontal' or 'vertical').
        min_threshold (float): Minimum threshold for shrinking. Default is 0.
        max_threshold (float): Maximum threshold for shrinking. Default is 0.2.

    Returns:
        list: List of tuples representing the merged intervals.
    """
    current_block = boxes[0]
    for block in boxes[1:]:
        x1, y1, x2, y2 = current_block.bbox
        x1_prime, y1_prime, x2_prime, y2_prime = block.bbox
        cut_iou = calculate_projection_overlap_ratio(
            current_block.bbox, block.bbox, direction=direction
        )
        match_iou = calculate_projection_overlap_ratio(
            current_block.bbox,
            block.bbox,
            direction="horizontal" if direction == "vertical" else "vertical",
        )
        if direction == "vertical":
            if (
                (match_iou > 0 and cut_iou > min_threshold and cut_iou < max_threshold)
                or y2 == y1_prime
                or abs(y2 - y1_prime) <= 3
            ):
                overlap_y_min = max(y1, y1_prime)
                overlap_y_max = min(y2, y2_prime)
                split_y = int((overlap_y_min + overlap_y_max) / 2)
                overlap_y_min = split_y - 1
                overlap_y_max = split_y + 1
                if y1 < y1_prime:
                    current_block.bbox = [x1, y1, x2, overlap_y_min]
                    block.bbox = [x1_prime, overlap_y_max, x2_prime, y2_prime]
                else:
                    current_block.bbox = [x1, overlap_y_min, x2, y2]
                    block.bbox = [x1_prime, y1_prime, x2_prime, overlap_y_max]
        else:
            if (
                (match_iou > 0 and cut_iou > min_threshold and cut_iou < max_threshold)
                or x2 == x1_prime
                or abs(x2 - x1_prime) <= 3
            ):
                overlap_x_min = max(x1, x1_prime)
                overlap_x_max = min(x2, x2_prime)
                split_x = int((overlap_x_min + overlap_x_max) / 2)
                overlap_x_min = split_x - 1
                overlap_x_max = split_x + 1
                if x1 < x1_prime:
                    current_block.bbox = [x1, y1, overlap_x_min, y2]
                    block.bbox = [overlap_x_max, y1_prime, x2_prime, y2_prime]
                else:
                    current_block.bbox = [overlap_x_min, y1, x2, y2]
                    block.bbox = [x1_prime, y1_prime, overlap_x_max, y2_prime]
        current_block = block
    return boxes


def find_local_minima_flat_regions(arr) -> List:
    """
    Find all local minima regions in a flat array.

    Args:
        arr (list): The input array.

    Returns:
        list: A list of tuples containing the indices of the local minima regions.
    """
    n = len(arr)
    if n == 0:
        return []

    flat_minima_regions = []
    start = 0

    for i in range(1, n):
        if arr[i] != arr[i - 1]:
            if (start == 0 or arr[start - 1] > arr[start]) and (
                i == n or arr[i] > arr[start]
            ):
                flat_minima_regions.append((start, i - 1))
            start = i

    return flat_minima_regions[1:] if len(flat_minima_regions) > 1 else None



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import subprocess

from ....utils.logging import info

__all__ = ["run_cmd", "CompletedProcess"]


def run_cmd(
    cmd,
    env=None,
    silent=True,
    cwd=None,
    timeout=None,
    echo=False,
    pipe_stdout=False,
    pipe_stderr=False,
    blocking=True,
    async_run=False,
    text=True,
):
    """Wrap around `subprocess.Popen` to execute a shell command."""
    # TODO: Limit argument length
    cfg = dict(env=env, cwd=cwd)

    async_run = async_run and not blocking

    if blocking:
        cfg["timeout"] = timeout
    if silent:
        cfg["stdout"] = (
            subprocess.DEVNULL if not async_run else asyncio.subprocess.DEVNULL
        )
        cfg["stderr"] = (
            subprocess.STDOUT if not async_run else asyncio.subprocess.STDOUT
        )
    if not async_run and (pipe_stdout or pipe_stderr):
        cfg["text"] = True
    if pipe_stdout:
        cfg["stdout"] = subprocess.PIPE if not async_run else asyncio.subprocess.PIPE
    if pipe_stderr:
        cfg["stderr"] = subprocess.PIPE if not async_run else asyncio.subprocess.PIPE

    if echo:
        info(str(cmd))

    if blocking:
        return subprocess.run(cmd, **cfg, check=False)
    else:
        if async_run:
            return asyncio.create_subprocess_exec(cmd[0], *cmd[1:], **cfg)
        else:
            if text:
                cfg.update(dict(bufsize=1, text=True))
            else:
                cfg.update(dict(bufsize=0, text=False))
            return subprocess.Popen(cmd, **cfg)


class CompletedProcess(object):
    """CompletedProcess"""

    __slots__ = ["args", "returncode", "stdout", "stderr", "_add_attrs"]

    def __init__(self, args, returncode, stdout=None, stderr=None):
        super().__init__()
        self.args = args
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self._add_attrs = dict()

    def __getattr__(self, name):
        try:
            val = self._add_attrs[name]
            return val
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, val):
        try:
            super().__setattr__(name, val)
        except AttributeError:
            self._add_attrs[name] = val

    def __repr__(self):
        args = [f"args={repr(self.args)}", f"returncode={repr(self.returncode)}"]
        if self.stdout is not None:
            args.append(f"stdout={repr(self.stdout)}")
        if self.stderr is not None:
            args.append(f"stderr={repr(self.stderr)}")
        return f"{self.__class__.__name__}({', '.join(args)})"
